from exegol.console.cli.ExegolCompleter import HybridContainerImageCompleter, VoidCompleter, BuildProfileCompleter
from exegol.console.cli.actions.Command import Command
from exegol.console.cli.actions.GenericParameters import *
from exegol.manager.ExegolManager import ExegolManager
from exegol.manager.UpdateManager import UpdateManager
from exegol.utils.ExeLog import logger


class Start(Command, ContainerCreation, ContainerSpawnShell):
    """Automatically create, start / resume and enter an Exegol container"""

    def __init__(self):
        Command.__init__(self)
        ContainerCreation.__init__(self, self.groupArgs)
        ContainerSpawnShell.__init__(self, self.groupArgs)

        self._usages = {
            "Start interactively a container": "exegol start",
            "Create a [blue]demo[/blue] container using [bright_blue]full[/bright_blue] image": "exegol start [blue]demo[/blue] [bright_blue]full[/bright_blue]",
            "Spawn a shell from [blue]demo[/blue] container": "exegol start [blue]demo[/blue]",
            "Create a container [blue]test[/blue] with a custom shared workspace": "exegol start [blue]test[/blue] [bright_blue]full[/bright_blue] -w [magenta]./project/pentest/[/magenta]",
            "Create a container [blue]test[/blue] sharing the current working directory": "exegol start [blue]test[/blue] [bright_blue]full[/bright_blue] -cwd",
            "Create a container [blue]htb[/blue] with a VPN": "exegol start [blue]htb[/blue] [bright_blue]full[/bright_blue] --vpn [magenta]~/vpn/[/magenta][bright_magenta]lab_Dramelac.ovpn[/bright_magenta]",
            "Create a container [blue]app[/blue] with custom volume": "exegol start [blue]app[/blue] [bright_blue]full[/bright_blue] -V [bright_magenta]'/var/app/:/app/'[/bright_magenta]",
            "Get a [blue]tmux[/blue] shell": "exegol start --shell [blue]tmux[/blue]",
            "Share a specific [blue]hardware device[/blue] (like Proxmark)": "exegol start -d /dev/ttyACM0",
            "Share every [blue]USB device[/blue] connected to the host": "exegol start -d /dev/bus/usb/",
        }

    def __call__(self, *args, **kwargs):
        return ExegolManager.start


class Stop(Command, ContainerMultiSelector):
    """Stop an Exegol container"""

    def __init__(self):
        Command.__init__(self)
        ContainerMultiSelector.__init__(self, self.groupArgs)

        self._usages = {
            "Stop interactively one or more containers": "exegol stop",
            "Stop [blue]demo[/blue]": "exegol stop [blue]demo[/blue]"
        }

    def __call__(self, *args, **kwargs):
        logger.debug("Running stop module")
        return ExegolManager.stop


class Restart(Command, ContainerSelector, ContainerSpawnShell):
    """Restart an Exegol container"""

    def __init__(self):
        Command.__init__(self)
        ContainerSelector.__init__(self, self.groupArgs)
        ContainerSpawnShell.__init__(self, self.groupArgs)

        self._usages = {
            "Restart interactively one containers": "exegol restart",
            "Restart [blue]demo[/blue]": "exegol restart [blue]demo[/blue]"
        }

    def __call__(self, *args, **kwargs):
        logger.debug("Running restart module")
        return ExegolManager.restart


class Install(Command, ImageSelector):
    """Install or build Exegol image"""

    def __init__(self):
        Command.__init__(self)
        ImageSelector.__init__(self, self.groupArgs)

        # Create container build arguments
        self.build_profile = Option("build_profile",
                                    metavar="BUILD_PROFILE",
                                    choices=UpdateManager.listBuildProfiles().keys(),
                                    nargs="?",
                                    action="store",
                                    help="Select the build profile used to create a local image.",
                                    completer=BuildProfileCompleter)
        self.build_log = Option("--build-log",
                                dest="build_log",
                                metavar="LOGFILE_PATH",
                                action="store",
                                help="Write image building logs to a file.")

        # Create group parameter for container selection
        self.groupArgs.append(GroupArg({"arg": self.build_profile, "required": False},
                                       {"arg": self.build_log, "required": False},
                                       title="[bold cyan]Build[/bold cyan] [blue]specific options[/blue]"))

        self._usages = {
            "Install or build interactively an exegol image": "exegol install",
            "Install or update the [bright_blue]full[/bright_blue] image": "exegol install [bright_blue]full[/bright_blue]",
            "Build interactively a local image named [blue]myimage[/blue]": "exegol install [blue]myimage[/blue]",
            "Build the [blue]myimage[/blue] image based on the [bright_blue]full[/bright_blue] profile and log the operation": "exegol install [blue]myimage[/blue] [bright_blue]full[/bright_blue] --build-log /tmp/build.log",
        }

    def __call__(self, *args, **kwargs):
        logger.debug("Running install module")
        return ExegolManager.install


class Update(Command, ImageSelector):
    """Update an Exegol image"""

    def __init__(self):
        Command.__init__(self)
        ImageSelector.__init__(self, self.groupArgs)

        self.skip_git = Option("--skip-git",
                               dest="skip_git",
                               action="store_true",
                               help="Skip git updates (wrapper, image sources and exegol resources).")
        self.skip_images = Option("--skip-images",
                                  dest="skip_images",
                                  action="store_true",
                                  help="Skip images updates (exegol docker images).")

        # Create group parameter for container selection
        self.groupArgs.append(GroupArg({"arg": self.skip_git, "required": False},
                                       {"arg": self.skip_images, "required": False},
                                       title="[bold cyan]Update[/bold cyan] [blue]specific options[/blue]"))

        self._usages = {
            "Install or update interactively an exegol image": "exegol update",
            "Install or update the [bright_blue]full[/bright_blue] image": "exegol update [bright_blue]full[/bright_blue]"
        }

    def __call__(self, *args, **kwargs):
        logger.debug("Running update module")
        return ExegolManager.update


class Uninstall(Command, ImageMultiSelector):
    """Remove Exegol [default not bold]image(s)[/default not bold]"""

    def __init__(self):
        Command.__init__(self)
        ImageMultiSelector.__init__(self, self.groupArgs)

        self.force_mode = Option("-F", "--force",
                                 dest="force_mode",
                                 action="store_true",
                                 help="Remove image without interactive user confirmation.")

        # Create group parameter for container selection
        self.groupArgs.append(GroupArg({"arg": self.force_mode, "required": False},
                                       title="[bold cyan]Uninstall[/bold cyan] [blue]specific options[/blue]"))

        self._usages = {
            "Uninstall interactively one or more exegol images": "exegol uninstall",
            "Uninstall the [bright_blue]dev[/bright_blue] image": "exegol uninstall [bright_blue]dev[/bright_blue]"
        }

    def __call__(self, *args, **kwargs):
        logger.debug("Running uninstall module")
        return ExegolManager.uninstall


class Remove(Command, ContainerMultiSelector):
    """Remove Exegol [default not bold]container(s)[/default not bold]"""

    def __init__(self):
        Command.__init__(self)
        ContainerMultiSelector.__init__(self, self.groupArgs)

        self.force_mode = Option("-F", "--force",
                                 dest="force_mode",
                                 action="store_true",
                                 help="Remove container without interactive user confirmation.")

        # Create group parameter for container selection
        self.groupArgs.append(GroupArg({"arg": self.force_mode, "required": False},
                                       title="[bold cyan]Remove[/bold cyan] [blue]specific options[/blue]"))

        self._usages = {
            "Remove interactively one or more containers": "exegol remove",
            "Remove the [blue]demo[/blue] container": "exegol remove [blue]demo[/blue]"
        }

    def __call__(self, *args, **kwargs):
        logger.debug("Running remove module")
        return ExegolManager.remove


class Exec(Command, ContainerCreation, ContainerStart):
    """Execute a command on an Exegol container"""

    def __init__(self):
        Command.__init__(self)
        ContainerCreation.__init__(self, self.groupArgs)
        ContainerStart.__init__(self, self.groupArgs)

        # Overwrite default selectors
        for group in self.groupArgs.copy():
            # Find group containing default selector to remove them
            for parameter in group.options:
                if parameter.get('arg') == self.containertag or parameter.get('arg') == self.imagetag:
                    # Removing default GroupArg selector
                    self.groupArgs.remove(group)
                    break
        # Removing default selector objects
        self.containertag = None
        self.imagetag = None

        self.selector = Option("selector",
                               metavar="CONTAINER or IMAGE",
                               nargs='?',
                               action="store",
                               help="Tag used to target an Exegol container (by default) or an image (if --tmp is set).",
                               completer=HybridContainerImageCompleter)

        # Custom parameters
        self.exec = Option("exec",
                           metavar="COMMAND",
                           nargs="+",
                           action="store",
                           help="Execute a single command in the exegol container.",
                           completer=VoidCompleter)
        self.daemon = Option("-b", "--background",
                             action="store_true",
                             dest="daemon",
                             help="Executes the command in background as a daemon "
                                  "(default: [red not italic]False[/red not italic])")
        self.tmp = Option("--tmp",
                          action="store_true",
                          dest="tmp",
                          help="Creates a dedicated and temporary container to execute the command "
                               "(default: [red not italic]False[/red not italic])")

        # Create group parameter for container selection
        self.groupArgs.append(GroupArg({"arg": self.selector, "required": False},
                                       {"arg": self.exec, "required": False},
                                       {"arg": self.daemon, "required": False},
                                       {"arg": self.tmp, "required": False},
                                       title="[bold cyan]Exec[/bold cyan] [blue]specific options[/blue]"))

        self._usages = {
            "Execute the command [magenta]bloodhound[/magenta] in the container [blue]demo[/blue]":
                "exegol exec [blue]demo[/blue] [magenta]bloodhound[/magenta]",
            "Execute the command [magenta]'nmap -h'[/magenta] with console output":
                "exegol exec -v [blue]demo[/blue] [magenta]'nmap -h'[/magenta]",
            "Execute a command in [green]background[/green] within the [blue]demo[/blue] container":
                "exegol exec [green]-b[/green] [blue]demo[/blue] [magenta]bloodhound[/magenta]",
            "Execute the command [magenta]bloodhound[/magenta] in a temporary container based on the [bright_blue]full[/bright_blue] image":
                "exegol exec --tmp [bright_blue]full[/bright_blue] [magenta]bloodhound[/magenta]",
            "Execute a command in [green]background[/green] with a temporary container":
                "exegol exec [green]-b[/green] --tmp [bright_blue]full[/bright_blue] [magenta]bloodhound[/magenta]",
            "Execute the command [magenta]wireshark[/magenta] with [orange3]network admin[/orange3] privileged":
                "exegol exec [green]-b[/green] --tmp --cap [orange3]NET_ADMIN[/orange3] [bright_blue]full[/bright_blue] [magenta]wireshark[/magenta]",
        }

    def __call__(self, *args, **kwargs):
        logger.debug("Running exec module")
        return ExegolManager.exec


class Info(Command, ContainerSelector):
    """Show info on containers and images (local & remote)"""

    def __init__(self):
        Command.__init__(self)
        ContainerSelector.__init__(self, self.groupArgs)

        self._usages = {
            "Print containers and images essentials information": "exegol info",
            "Print the detailed configuration of the [blue]demo[/blue] container": "exegol info [blue]demo[/blue]",
            "Print verbose information": "exegol info [yellow3]-v[/yellow3]",
            "Print advanced information": "exegol info [yellow3]-vv[/yellow3]",
            "Print debug information": "exegol info [yellow3]-vvv[/yellow3]"
        }

    def __call__(self, *args, **kwargs):
        return ExegolManager.info


class Version(Command):
    """Print current Exegol version"""

    def __call__(self, *args, **kwargs):
        return ExegolManager.print_version



import rich.prompt


def Confirm(question: str, default: bool) -> bool:
    """Quick function to format rich Confirmation and options on every exegol interaction"""
    default_text = "[bright_magenta][Y/n][/bright_magenta]" if default else "[bright_magenta]\[y/N][/bright_magenta]"
    formatted_question = f"[bold blue][?][/bold blue] {question} {default_text}"
    return rich.prompt.Confirm.ask(
        formatted_question,
        show_choices=False,
        show_default=False,
        default=default)



import subprocess
import threading
import tempfile

# The file containing the list of commands
commands_file = "/.exegol/build_pipeline_tests/all_commands.sorted.txt"

# The file to store the logs of commands
fail_log_file = "/.exegol/build_pipeline_tests/failed_commands.log"
success_log_file = "/.exegol/build_pipeline_tests/success_commands.log"

# Read the commands from the file
with open(commands_file) as f:
    commands = f.readlines()

# Remove the newline characters from the commands
commands = [x.strip() for x in commands]

# Initialize the variable to store the failed commands
failed_commands = []


# Define a function to run a single command
def run_command(command):
    # Create a temporary file for the command
    with tempfile.NamedTemporaryFile(mode="w") as temp:
        # Write the command to the temporary file
        temp.write(command)
        temp.flush()

        try:
            # Try to run the command in a zsh context
            zsh_command = f"zsh -c 'autoload -Uz compinit; compinit; source ~/.zshrc; . {temp.name}'"
            output = subprocess.check_output(zsh_command, shell=True, stderr=subprocess.PIPE, timeout=120)
            print(f"\033[1;32mSUCCESS\033[0m - Running command: {command}")

            # Write the output of the successful command to the log file
            with open(success_log_file, "a") as f:
                f.write(f"\033[1;34mCommand: {command}\n\033[0m")
                f.write("\033[33mStandard output:\n")
                for line in output.decode().split("\n"):
                    f.write(f"    {line}\n")
                f.write("\033[0m")
        except subprocess.TimeoutExpired as e:
            # If the command timeout, store it in the list of failed commands
            failed_commands.append(command)
            print(f"\033[1;31mTIMEOUT\033[0m - Running command: {command}")
        except subprocess.CalledProcessError as e:
            # If the command fails, store it in the list of failed commands
            failed_commands.append(command)
            print(f"\033[1;31mFAILURE\033[0m - Running command: {command}")

            # Write the output of the failed command to the log file
            with open(fail_log_file, "a") as f:
                f.write(f"\033[1;34mFailed command: {command}\n\033[0m")
                if e.output:
                    f.write("\033[33mStandard output:\n")
                    for line in e.output.decode().split("\n"):
                        f.write(f"    {line}\n")
                    f.write("\033[0m")
                f.write("\033[31mStandard error:\n")
                for line in e.stderr.decode().split("\n"):
                    f.write(f"    {line}\n")
                f.write("\033[0m")


# Create a list of threads
threads = [threading.Thread(target=run_command, args=(command,)) for command in commands]

# Start the threads
for thread in threads:
    thread.start()

# Wait for the threads to finish
for thread in threads:
    thread.join()

# Check if any of the commands failed
if failed_commands:
    print("\033[33mThe following commands failed:\033[0m")
    for command in failed_commands:
        print(f"    {command}")
    print(f"\033[33mLogs of failed commands are stored in\033[0m {fail_log_file}")
    print(f"\033[33mLogs of success commands are stored in\033[0m {success_log_file}")
    exit(1)
else:
    print("All commands succeeded.")
    print(f"\033[33mLogs of success commands are stored in\033[0m {success_log_file}")
    exit(0)



import site
from pathlib import Path


class ConstantConfig:
    """Constant parameters information"""
    # Exegol Version
    version: str = "4.2.5"

    # Exegol documentation link
    documentation: str = "https://exegol.rtfd.io/"
    discord: str = "https://discord.gg/cXThyp7D6P"
    # OS Dir full root path of exegol project
    src_root_path_obj: Path = Path(__file__).parent.parent.parent.resolve()
    # Path of the Dockerfile
    build_context_path_obj: Path
    build_context_path: str
    # Exegol config directory
    exegol_config_path: Path = Path().home() / ".exegol"
    # Docker Desktop for mac config file
    docker_desktop_mac_config_path = Path().home() / "Library/Group Containers/group.com.docker/settings.json"
    docker_desktop_windows_config_path = Path().home() / "AppData/Roaming/Docker/settings.json"
    # Install mode, check if Exegol has been git cloned or installed using pip package
    git_source_installation: bool = (src_root_path_obj / '.git').is_dir()
    pip_installed: bool = src_root_path_obj.name == "site-packages"
    # Dockerhub Exegol images repository
    DOCKER_HUB: str = "hub.docker.com"  # Don't handle docker login operations
    DOCKER_REGISTRY: str = "registry-1.docker.io"  # Don't handle docker login operations
    IMAGE_NAME: str = "nwodtuhs/exegol"
    GITHUB_REPO: str = "ThePorgs/Exegol"
    # Docker volume names (no docker volume used at this moment)
    # Resources repository
    EXEGOL_RESOURCES_REPO: str = "https://github.com/ThePorgs/Exegol-resources.git"

    @classmethod
    def findBuildContextPath(cls) -> Path:
        """Find the right path to the build context from Exegol docker images.
        Support source clone installation and pip package (venv / user / global context)"""
        dockerbuild_folder_name = "exegol-docker-build"
        local_src = cls.src_root_path_obj / dockerbuild_folder_name
        if local_src.is_dir():
            # If exegol is clone from github, build context is accessible from root src
            return local_src
        else:
            # If install from pip
            if site.ENABLE_USER_SITE:
                # Detect a user based python env
                possible_locations = [Path(site.getuserbase())]
                # Detect a global installed package
                for loc in site.getsitepackages():
                    possible_locations.append(Path(loc).parent.parent.parent)
                # Find a good match
                for test in possible_locations:
                    context_path = test / dockerbuild_folder_name
                    if context_path.is_dir():
                        return context_path
            # Detect a venv context
            return Path(site.PREFIXES[0]) / dockerbuild_folder_name


# Dynamically built attribute must be set after class initialization
ConstantConfig.build_context_path_obj = ConstantConfig.findBuildContextPath()
ConstantConfig.build_context_path = str(ConstantConfig.build_context_path_obj)



try:
    from git.exc import GitCommandError

    from exegol.console.cli.ParametersManager import ParametersManager
    from exegol.console.cli.actions.ExegolParameters import Command
    from exegol.utils.ExeLog import logger, ExeLog, console
except ModuleNotFoundError as e:
    print("Mandatory dependencies are missing:", e)
    print("Please install them with python3 -m pip install --upgrade -r requirements.txt")
    exit(1)
except ImportError as e:
    print("An error occurred while loading the dependencies!")
    print()
    if "git executable" in e.msg:
        print("Git is missing in your PATH, it must be installed locally on your computer.")
        print()
    print("Details:")
    print(e)
    exit(1)


class ExegolController:
    """Main controller of exegol"""

    # Get action selected by user
    # (ParametersManager must be loaded from ExegolController first to load every Command subclass)
    __action: Command = ParametersManager().getCurrentAction()

    @classmethod
    def call_action(cls):
        """Dynamically retrieve the main function corresponding to the action selected by the user
        and execute it on the main thread"""
        # Check for missing parameters
        missing_params = cls.__action.check_parameters()
        if len(missing_params) == 0:
            # Fetch main operation function
            main_action = cls.__action()
            # Execute main function
            main_action()
        else:
            # TODO review required parameters
            logger.error(f"These parameters are mandatory but missing: {','.join(missing_params)}")


def print_exception_banner():
    logger.error("It seems that something unexpected happened ...")
    logger.error("To draw our attention to the problem and allow us to fix it, you can share your error with us "
                 "(by [orange3]copying and pasting[/orange3] it with this syntax: ``` <error> ```) "
                 "by creating a GitHub issue at this address: https://github.com/ThePorgs/Exegol/issues")
    logger.success("Thank you for your collaboration!")


def main():
    """Exegol main console entrypoint"""
    try:
        # Set logger verbosity depending on user input
        ExeLog.setVerbosity(ParametersManager().verbosity, ParametersManager().quiet)
        # Start Main controller & Executing action selected by user CLI
        ExegolController.call_action()
    except KeyboardInterrupt:
        logger.empty_line()
        logger.info("Exiting")
    except GitCommandError as e:
        print_exception_banner()
        error = e.stderr.strip().split(": ")[-1].strip("'")
        logger.critical(f"A critical error occurred while running this git command: {' '.join(e.command)} => {error}")
    except Exception:
        print_exception_banner()
        console.print_exception(show_locals=True)
        exit(1)



from typing import Optional, Set, Union

from docker.models.images import Image

from exegol.utils.ExeLog import logger
from exegol.utils.WebUtils import WebUtils


class MetaImages:
    """Meta model to store and organise multi-arch images"""

    def __init__(self, dockerhub_data):
        """Create a MetaImage object to handle multi-arch docker registry images in a single point"""
        # Raw data
        self.__dockerhub_images: dict = dockerhub_data.get('images')
        # Attributes
        self.name: str = dockerhub_data.get('name', '')
        self.multi_arch: bool = len(self.__dockerhub_images) > 1
        self.list_arch: Set[str] = set(
            [self.parseArch(a) for a in self.__dockerhub_images])
        self.meta_id: Optional[str] = dockerhub_data.get("digest")
        if not self.meta_id:
            if self.multi_arch:
                logger.debug(f"Missing ID for image {self.name}, manual fetching ! May slow down the process..")
                self.meta_id = WebUtils.getMetaDigestId(self.name)
            else:
                # Single arch image dont need virtual meta_id
                self.__dockerhub_images[0].get('digest')
        self.version: str = self.tagNameParsing(self.name)
        self.is_latest: bool = not bool(self.version)  # Current image is latest if no version have been found from tag name
        # Post-process data
        self.__image_match = set()

    @staticmethod
    def tagNameParsing(tag_name: str) -> str:
        parts = tag_name.split('-')
        version = '-'.join(parts[1:])
        # Code for future multi parameter from tag name (e.g. ad-debian-1.2.3)
        """
        first_parameter = ""
        # Try to detect legacy tag name or new latest name
        if len(parts) == 2:
            # If there is any '.' in the second part, it's a version format
            if "." in parts[1]:
                # Legacy version format
                version = parts[1]
            else:
                # Latest arch specific image
                first_parameter = parts[1]
        elif len(parts) >= 3:
            # Arch + version format
            first_parameter = parts[1]
            # Additional - stored in version
            version = '-'.join(parts[2:])

        return version, first_parameter
        """
        return version

    @staticmethod
    def parseArch(docker_image: Union[dict, Image]) -> str:
        """Parse and format arch in dockerhub style from registry dict struct.
        Return arch in format 'arch/variant'."""
        arch_key = "architecture"
        variant_key = "variant"
        # Support Docker image struct with specific dict key
        if type(docker_image) is Image:
            docker_image = docker_image.attrs
            arch_key = "Architecture"
            variant_key = "Variant"
        arch = docker_image.get(arch_key, "amd64")
        variant = docker_image.get(variant_key)
        if variant:
            arch += f"/{variant}"
        return arch

    def getDockerhubImageForArch(self, arch: str) -> Optional[dict]:
        """Find a docker image corresponding to a specific arch"""
        for img in self.__dockerhub_images:
            if self.parseArch(img) == arch:
                self.__image_match.add(arch)
                return img
        return None

    def getImagesLeft(self):
        """Return every image not previously selected."""
        result = []
        for img in self.__dockerhub_images:
            if self.parseArch(img) not in self.__image_match:
                result.append(img)
        return result

    def setVersionSpecific(self, meta_version: 'MetaImages'):
        self.version = meta_version.version

    def __str__(self):
        return f"{self.name} ({self.version}) [{self.meta_id}] {self.list_arch}"

    def __repr__(self):
        return self.__str__()



import logging
import os
import re
from datetime import datetime
from pathlib import Path, PurePath
from typing import Optional, List, Dict, Union, Tuple, cast

from docker.models.containers import Container
from docker.types import Mount
from rich.prompt import Prompt

from exegol.console.ConsoleFormat import boolFormatter, getColor
from exegol.console.ExegolPrompt import Confirm
from exegol.console.cli.ParametersManager import ParametersManager
from exegol.exceptions.ExegolExceptions import ProtocolNotSupported, CancelOperation
from exegol.model.ExegolModules import ExegolModules
from exegol.utils import FsUtils
from exegol.config.EnvInfo import EnvInfo
from exegol.utils.ExeLog import logger, ExeLog
from exegol.utils.GuiUtils import GuiUtils
from exegol.config.UserConfig import UserConfig


class ContainerConfig:
    """Configuration class of an exegol container"""

    # Default hardcoded value
    __default_entrypoint_legacy = "bash"
    __default_entrypoint = ["/.exegol/entrypoint.sh"]
    __default_cmd = ["default"]
    __default_shm_size = "64M"

    # Reference static config data
    __static_gui_envs = {"_JAVA_AWT_WM_NONREPARENTING": "1", "QT_X11_NO_MITSHM": "1"}

    # Label features (wrapper method to enable the feature / label name)
    __label_features = {"enableShellLogging": "org.exegol.feature.shell_logging"}
    # Label metadata (label name / [wrapper attribute to set the value, getter method to update labels])
    __label_metadata = {"org.exegol.metadata.creation_date": ["creation_date", "getCreationDate"],
                        "org.exegol.metadata.comment": ["comment", "getComment"]}

    def __init__(self, container: Optional[Container] = None):
        """Container config default value"""
        self.__enable_gui: bool = False
        self.__share_timezone: bool = False
        self.__my_resources: bool = False
        self.__my_resources_path: str = "/opt/my-resources"
        self.__exegol_resources: bool = False
        self.__network_host: bool = True
        self.__privileged: bool = False
        self.__mounts: List[Mount] = []
        self.__devices: List[str] = []
        self.__capabilities: List[str] = []
        self.__sysctls: Dict[str, str] = {}
        self.__envs: Dict[str, str] = {}
        self.__labels: Dict[str, str] = {}
        self.__ports: Dict[str, Optional[Union[int, Tuple[str, int], List[int], List[Dict[str, Union[int, str]]]]]] = {}
        self.interactive: bool = True
        self.tty: bool = True
        self.shm_size: str = self.__default_shm_size
        self.__workspace_custom_path: Optional[str] = None
        self.__workspace_dedicated_path: Optional[str] = None
        self.__disable_workspace: bool = False
        self.__container_command_legacy: Optional[str] = None
        self.__container_command: List[str] = self.__default_cmd
        self.__container_entrypoint: List[str] = self.__default_entrypoint
        self.__vpn_path: Optional[Union[Path, PurePath]] = None
        self.__shell_logging: bool = False
        self.__start_delegate_mode: bool = False
        # Metadata attributes
        self.creation_date: Optional[str] = None
        self.comment: Optional[str] = None

        if container is not None:
            self.__parseContainerConfig(container)

    def __parseContainerConfig(self, container: Container):
        """Parse Docker object to setup self configuration"""
        # Container Config section
        container_config = container.attrs.get("Config", {})
        self.tty = container_config.get("Tty", True)
        self.__parseEnvs(container_config.get("Env", []))
        self.__parseLabels(container_config.get("Labels", {}))
        self.interactive = container_config.get("OpenStdin", True)
        # If entrypoint is set on the image, considering the presence of start.sh script for delegates features
        self.__start_delegate_mode = container.attrs['Config']['Entrypoint'] is not None
        self.__enable_gui = False
        for env in self.__envs:
            if "DISPLAY" in env:
                self.__enable_gui = True
                break

        # Host Config section
        host_config = container.attrs.get("HostConfig", {})
        self.__privileged = host_config.get("Privileged", False)
        caps = host_config.get("CapAdd", [])
        if caps is not None:
            self.__capabilities = caps
        logger.debug(f"Capabilities : {self.__capabilities}")
        self.__sysctls = host_config.get("Sysctls", {})
        devices = host_config.get("Devices", [])
        if devices is not None:
            for device in devices:
                self.__devices.append(
                    f"{device.get('PathOnHost', '?')}:{device.get('PathInContainer', '?')}:{device.get('CgroupPermissions', '?')}")
        logger.debug(f"Load devices : {self.__devices}")

        # Volumes section
        self.__share_timezone = False
        self.__my_resources = False
        self.__parseMounts(container.attrs.get("Mounts", []), container.name.replace('exegol-', ''))

        # Network section
        network_settings = container.attrs.get("NetworkSettings", {})
        self.__network_host = "host" in network_settings["Networks"]
        self.__ports = network_settings.get("Ports", {})

    def __parseEnvs(self, envs: List[str]):
        """Parse envs object syntax"""
        for env in envs:
            logger.debug(f"Parsing envs : {env}")
            # Removing " and ' at the beginning and the end of the string before splitting key / value
            self.addRawEnv(env.strip("'").strip('"'))

    def __parseLabels(self, labels: Dict[str, str]):
        """Parse envs object syntax"""
        for key, value in labels.items():
            if not key.startswith("org.exegol."):
                continue
            logger.debug(f"Parsing label : {key}")
            if key.startswith("org.exegol.metadata."):
                # Find corresponding feature and attributes
                for label, refs in self.__label_metadata.items():
                    if label == key:
                        # reflective set of the metadata attribute (set metadata value to the corresponding attribute)
                        setattr(self, refs[0], value)
                        break
            elif key.startswith("org.exegol.feature."):
                # Find corresponding feature and attributes
                for attribute, label in self.__label_features.items():
                    if label == key:
                        # reflective execution of the feature enable method (add label & set attributes)
                        getattr(self, attribute)()
                        break

    def __parseMounts(self, mounts: Optional[List[Dict]], name: str):
        """Parse Mounts object"""
        if mounts is None:
            mounts = []
        self.__disable_workspace = True
        for share in mounts:
            logger.debug(f"Parsing mount : {share}")
            src_path: Optional[PurePath] = None
            obj_path: PurePath
            if share.get('Type', 'volume') == "volume":
                source = f"Docker {share.get('Driver', '')} volume '{share.get('Name', 'unknown')}'"
            else:
                source = share.get("Source", '')
                src_path = FsUtils.parseDockerVolumePath(source)

                # When debug is disabled, exegol print resolved windows path of mounts
                if logger.getEffectiveLevel() > ExeLog.ADVANCED:
                    source = str(src_path)

            self.__mounts.append(Mount(source=source,
                                       target=share.get('Destination'),
                                       type=share.get('Type', 'volume'),
                                       read_only=(not share.get("RW", True)),
                                       propagation=share.get('Propagation', '')))
            if share.get('Destination', '') in ["/etc/timezone", "/etc/localtime"]:
                self.__share_timezone = True
            elif "/opt/resources" in share.get('Destination', ''):
                self.__exegol_resources = True
            # TODO remove support for previous container
            elif "/my-resources" in share.get('Destination', '') or "/opt/my-resources" in share.get('Destination', ''):
                self.__my_resources = True
                self.__my_resources_path = share.get('Destination', '')
            elif "/workspace" in share.get('Destination', ''):
                # Workspace are always bind mount
                assert src_path is not None
                obj_path = cast(PurePath, src_path)
                logger.debug(f"Loading workspace volume source : {obj_path}")
                self.__disable_workspace = False
                if obj_path is not None and obj_path.name == name and \
                        (obj_path.parent.name == "shared-data-volumes" or obj_path.parent == UserConfig().private_volume_path):  # Check legacy path and new custom path
                    logger.debug("Private workspace detected")
                    self.__workspace_dedicated_path = str(obj_path)
                else:
                    logger.debug("Custom workspace detected")
                    self.__workspace_custom_path = str(obj_path)
            # TODO remove support for previous container
            elif "/vpn" in share.get('Destination', '') or "/.exegol/vpn" in share.get('Destination', ''):
                # VPN are always bind mount
                assert src_path is not None
                obj_path = cast(PurePath, src_path)
                self.__vpn_path = obj_path
                logger.debug(f"Loading VPN config: {self.__vpn_path.name}")

    def interactiveConfig(self, container_name: str) -> List[str]:
        """Interactive procedure allowing the user to configure its new container"""
        logger.info("Starting interactive configuration")

        command_options = []

        # Workspace config
        if Confirm(
                "Do you want to [green]share[/green] your [blue]current host working directory[/blue] in the new container's worskpace?",
                default=False):
            self.enableCwdShare()
            command_options.append("-cwd")
        elif Confirm(
                f"Do you want to [green]share[/green] [blue]a host directory[/blue] in the new container's workspace [blue]different than the default one[/blue] ([magenta]{UserConfig().private_volume_path / container_name}[/magenta])?",
                default=False):
            while True:
                workspace_path = Prompt.ask("Enter the path of your workspace")
                if Path(workspace_path).expanduser().is_dir():
                    break
                else:
                    logger.error("The provided path is not a folder or does not exist.")
            self.setWorkspaceShare(workspace_path)
            command_options.append(f"-w {workspace_path}")

        # GUI Config
        if self.__enable_gui:
            if Confirm("Do you want to [orange3]disable[/orange3] [blue]GUI[/blue]?", False):
                self.__disableGUI()
        elif Confirm("Do you want to [green]enable[/green] [blue]GUI[/blue]?", False):
            self.enableGUI()
        # Command builder info
        if not self.__enable_gui:
            command_options.append("--disable-X11")

        # Timezone config
        if self.__share_timezone:
            if Confirm("Do you want to [orange3]remove[/orange3] your [blue]shared timezone[/blue] config?", False):
                self.__disableSharedTimezone()
        elif Confirm("Do you want to [green]share[/green] your [blue]host's timezone[/blue]?", False):
            self.enableSharedTimezone()
        # Command builder info
        if not self.__share_timezone:
            command_options.append("--disable-shared-timezones")

        # my-resources config
        if self.__my_resources:
            if Confirm("Do you want to [orange3]disable[/orange3] [blue]my-resources[/blue]?", False):
                self.__disableMyResources()
        elif Confirm("Do you want to [green]activate[/green] [blue]my-resources[/blue]?", False):
            self.enableMyResources()
        # Command builder info
        if not self.__my_resources:
            command_options.append("--disable-my-resources")

        # Exegol resources config
        if self.__exegol_resources:
            if Confirm("Do you want to [orange3]disable[/orange3] the [blue]exegol resources[/blue]?", False):
                self.disableExegolResources()
        elif Confirm("Do you want to [green]activate[/green] the [blue]exegol resources[/blue]?", False):
            self.enableExegolResources()
        # Command builder info
        if not self.__exegol_resources:
            command_options.append("--disable-exegol-resources")

        # Network config
        if self.__network_host:
            if Confirm("Do you want to use a [blue]dedicated private network[/blue]?", False):
                self.setNetworkMode(False)
        elif Confirm("Do you want to share the [green]host's[/green] [blue]networks[/blue]?", False):
            self.setNetworkMode(True)
        # Command builder info
        if not self.__network_host:
            command_options.append("--disable-shared-network")

        # Shell logging config
        if self.__shell_logging:
            if Confirm("Do you want to [orange3]disable[/orange3] automatic [blue]shell logging[/blue]?", False):
                self.__disableShellLogging()
        elif Confirm("Do you want to [green]enable[/green] automatic [blue]shell logging[/blue]?", False):
            self.enableShellLogging()
        # Command builder info
        if self.__shell_logging:
            command_options.append("--log")

        # VPN config
        if self.__vpn_path is None and Confirm(
                "Do you want to [green]enable[/green] a [blue]VPN[/blue] for this container", False):
            while True:
                vpn_path = Prompt.ask('Enter the path to the OpenVPN config file')
                if Path(vpn_path).expanduser().is_file():
                    self.enableVPN(vpn_path)
                    break
                else:
                    logger.error("No config files were found.")
        elif self.__vpn_path and Confirm(
                "Do you want to [orange3]remove[/orange3] your [blue]VPN configuration[/blue] in this container", False):
            self.__disableVPN()
        if self.__vpn_path:
            command_options.append(f"--vpn {self.__vpn_path}")

        return command_options

    def enableGUI(self):
        """Procedure to enable GUI feature"""
        if not GuiUtils.isGuiAvailable():
            logger.error("GUI feature is [red]not available[/red] on your environment. [orange3]Skipping[/orange3].")
            return
        if not self.__enable_gui:
            logger.verbose("Config: Enabling display sharing")
            try:
                host_path = GuiUtils.getX11SocketPath()
                if host_path is not None:
                    self.addVolume(host_path, GuiUtils.default_x11_path, must_exist=True)
            except CancelOperation as e:
                logger.warning(f"Graphical interface sharing could not be enabled: {e}")
                return
            # TODO support pulseaudio
            self.addEnv("DISPLAY", GuiUtils.getDisplayEnv())
            for k, v in self.__static_gui_envs.items():
                self.addEnv(k, v)
            self.__enable_gui = True

    def __disableGUI(self):
        """Procedure to enable GUI feature (Only for interactive config)"""
        if self.__enable_gui:
            self.__enable_gui = False
            logger.verbose("Config: Disabling display sharing")
            self.removeVolume(container_path="/tmp/.X11-unix")
            self.removeEnv("DISPLAY")
            for k in self.__static_gui_envs.keys():
                self.removeEnv(k)

    def enableSharedTimezone(self):
        """Procedure to enable shared timezone feature"""
        if EnvInfo.is_windows_shell:
            logger.warning("Timezone sharing is not supported from a Windows shell. Skipping.")
            return
        if not self.__share_timezone:
            logger.verbose("Config: Enabling host timezones")
            # Try to share /etc/timezone (deprecated old timezone file)
            try:
                self.addVolume("/etc/timezone", "/etc/timezone", read_only=True, must_exist=True)
                logger.verbose("Volume was successfully added for [magenta]/etc/timezone[/magenta]")
                timezone_loaded = True
            except CancelOperation:
                logger.verbose("File /etc/timezone is missing on host, cannot create volume for this.")
                timezone_loaded = False
            # Try to share /etc/localtime (new timezone file)
            try:
                self.addVolume("/etc/localtime", "/etc/localtime", read_only=True, must_exist=True)
                logger.verbose("Volume was successfully added for [magenta]/etc/localtime[/magenta]")
            except CancelOperation as e:
                if not timezone_loaded:
                    # If neither file was found, disable the functionality
                    logger.error(f"The host's timezone could not be shared: {e}")
                    return
                else:
                    logger.warning("File [magenta]/etc/localtime[/magenta] is [orange3]missing[/orange3] on host, "
                                   "cannot create volume for this. Relying instead on [magenta]/etc/timezone[/magenta] [orange3](deprecated)[/orange3].")
            self.__share_timezone = True

    def __disableSharedTimezone(self):
        """Procedure to disable shared timezone feature (Only for interactive config)"""
        if self.__share_timezone:
            self.__share_timezone = False
            logger.verbose("Config: Disabling host timezones")
            self.removeVolume("/etc/timezone")
            self.removeVolume("/etc/localtime")

    def setPrivileged(self, status: bool = True):
        """Set container as privileged"""
        logger.verbose(f"Config: Setting container privileged as {status}")
        if status:
            logger.warning("Setting container as privileged (this exposes the host to security risks)")
        self.__privileged = status

    def enableMyResources(self):
        """Procedure to enable shared volume feature"""
        # TODO test my resources cross shell source (WSL / PSH) on Windows
        if not self.__my_resources:
            logger.verbose("Config: Enabling my-resources volume")
            self.__my_resources = True
            # Adding volume config
            self.addVolume(str(UserConfig().my_resources_path), '/opt/my-resources', enable_sticky_group=True, force_sticky_group=True)

    def __disableMyResources(self):
        """Procedure to disable shared volume feature (Only for interactive config)"""
        if self.__my_resources:
            logger.verbose("Config: Disabling my-resources volume")
            self.__my_resources = False
            self.removeVolume(container_path='/opt/my-resources')

    def enableExegolResources(self) -> bool:
        """Procedure to enable exegol resources volume feature"""
        if not self.__exegol_resources:
            # Check if resources are installed / up-to-date
            try:
                if not ExegolModules().isExegolResourcesReady():
                    raise CancelOperation
            except CancelOperation:
                # Error during installation, skipping operation
                logger.warning("Exegol resources have not been downloaded, the feature cannot be enabled")
                return False
            logger.verbose("Config: Enabling exegol resources volume")
            self.__exegol_resources = True
            # Adding volume config
            self.addVolume(str(UserConfig().exegol_resources_path), '/opt/resources')
        return True

    def disableExegolResources(self):
        """Procedure to disable exegol resources volume feature (Only for interactive config)"""
        if self.__exegol_resources:
            logger.verbose("Config: Disabling exegol resources volume")
            self.__exegol_resources = False
            self.removeVolume(container_path='/opt/resources')

    def enableShellLogging(self):
        """Procedure to enable exegol shell logging feature"""
        if not self.__shell_logging:
            logger.verbose("Config: Enabling shell logging")
            self.__shell_logging = True
            self.addLabel(self.__label_features.get('enableShellLogging', 'org.exegol.error'), "Enabled")

    def addComment(self, comment):
        """Procedure to add comment to a container"""
        if not self.comment:
            logger.verbose("Config: Adding comment to container info")
            self.comment = comment
            self.addLabel("org.exegol.metadata.comment", comment)

    def __disableShellLogging(self):
        """Procedure to disable exegol shell logging feature"""
        if self.__shell_logging:
            logger.verbose("Config: Disabling shell logging")
            self.__shell_logging = False
            self.removeLabel(self.__label_features.get('enableShellLogging', 'org.exegol.error'))

    def enableCwdShare(self):
        """Procedure to share Current Working Directory with the /workspace of the container"""
        self.__workspace_custom_path = os.getcwd()
        logger.verbose(f"Config: Sharing current workspace directory {self.__workspace_custom_path}")

    def setWorkspaceShare(self, host_directory):
        """Procedure to share a specific directory with the /workspace of the container"""
        path = Path(host_directory).expanduser().absolute()
        try:
            if not path.is_dir() and path.exists():
                logger.critical("The specified workspace is not a directory!")
        except PermissionError as e:
            logger.critical(f"Unable to use the supplied workspace directory: {e}")
        logger.verbose(f"Config: Sharing workspace directory {path}")
        self.__workspace_custom_path = str(path)

    def enableVPN(self, config_path: Optional[str] = None):
        """Configure a VPN profile for container startup"""
        # Check host mode : custom (allows you to isolate the VPN connection from the host's network)
        if self.__network_host:
            logger.warning("Using the host network mode with a VPN profile is not recommended.")
            if not Confirm(f"Are you sure you want to configure a VPN container based on the host's network?",
                           default=False):
                logger.info("Changing network mode to custom")
                self.setNetworkMode(False)
        # Add NET_ADMIN capabilities, this privilege is necessary to mount network tunnels
        self.addCapability("NET_ADMIN")
        # Add sysctl ipv6 config, some VPN connection need IPv6 to be enabled
        # TODO test with ipv6 disable with kernel modules
        skip_sysctl = False
        if self.__network_host and EnvInfo.is_linux_shell:
            # Check if IPv6 have been disabled on the host with sysctl
            with open('/proc/sys/net/ipv6/conf/all/disable_ipv6', 'r') as conf:
                if int(conf.read()) == 0:
                    skip_sysctl = True
        if not skip_sysctl:
            self.__addSysctl("net.ipv6.conf.all.disable_ipv6", "0")
        # Add tun device, this device is needed to create VPN tunnels
        self.__addDevice("/dev/net/tun", mknod=True)
        # Sharing VPN configuration with the container
        ovpn_parameters = self.__prepareVpnVolumes(config_path)
        # Execution of the VPN daemon at container startup
        if ovpn_parameters is not None:
            vpn_cmd_legacy = f"bash -c 'mkdir -p /var/log/exegol; openvpn --log-append /var/log/exegol/vpn.log {ovpn_parameters}; bash'"
            self.setLegacyContainerCommand(vpn_cmd_legacy)
            self.setContainerCommand("ovpn", ovpn_parameters)

    def __prepareVpnVolumes(self, config_path: Optional[str]) -> Optional[str]:
        """Volumes must be prepared to share OpenVPN configuration files with the container.
        Depending on the user's settings, different configurations can be applied.
        With or without username / password authentication via auth-user-pass.
        OVPN config file directly supplied or a config directory,
        the directory feature is useful when the configuration depends on multiple files like certificate, keys etc."""
        ovpn_parameters = []

        # VPN config path
        vpn_path = Path(config_path if config_path else ParametersManager().vpn).expanduser()

        logger.debug(f"Adding VPN from: {str(vpn_path.absolute())}")
        self.__vpn_path = vpn_path
        if vpn_path.is_file():
            self.__checkVPNConfigDNS(vpn_path)
            # Configure VPN with single file
            self.addVolume(str(vpn_path.absolute()), "/.exegol/vpn/config/client.ovpn", read_only=True)
            ovpn_parameters.append("--config /.exegol/vpn/config/client.ovpn")
        else:
            # Configure VPN with directory
            logger.verbose("Folder detected for VPN configuration. "
                           "Only the first *.ovpn file will be automatically launched when the container starts.")
            self.addVolume(str(vpn_path.absolute()), "/.exegol/vpn/config", read_only=True)
            vpn_filename = None
            # Try to find the config file in order to configure the autostart command of the container
            for file in vpn_path.glob('*.ovpn'):
                logger.info(f"Using VPN config: {file}")
                self.__checkVPNConfigDNS(file)
                # Get filename only to match the future container path
                vpn_filename = file.name
                ovpn_parameters.append(f"--config /.exegol/vpn/config/{vpn_filename}")
                # If there is multiple match, only the first one is selected
                break
            if vpn_filename is None:
                logger.error("No *.ovpn files were detected. The VPN autostart will not work.")
                return None

        # VPN Auth creds file
        input_vpn_auth = ParametersManager().vpn_auth
        vpn_auth = None
        if input_vpn_auth is not None:
            vpn_auth = Path(input_vpn_auth).expanduser()

        if vpn_auth is not None:
            if vpn_auth.is_file():
                logger.info(f"Adding VPN credentials from: {str(vpn_auth.absolute())}")
                self.addVolume(str(vpn_auth.absolute()), "/.exegol/vpn/auth/creds.txt", read_only=True)
                ovpn_parameters.append("--auth-user-pass /.exegol/vpn/auth/creds.txt")
            else:
                # Supply a directory instead of a file for VPN authentication is not supported.
                logger.critical(
                    f"The path provided to the VPN connection credentials ({str(vpn_auth)}) does not lead to a file. Aborting operation.")

        return ' '.join(ovpn_parameters)

    @staticmethod
    def __checkVPNConfigDNS(vpn_path: Union[str, Path]):
        logger.verbose("Checking OpenVPN config file")
        configs = ["script-security 2", "up /etc/openvpn/update-resolv-conf", "down /etc/openvpn/update-resolv-conf"]
        with open(vpn_path, 'r') as vpn_file:
            for line in vpn_file:
                line = line.strip()
                if line in configs:
                    configs.remove(line)
        if len(configs) > 0:
            logger.warning("Some OpenVPN config are [red]missing[/red] to support VPN [orange3]dynamic DNS servers[/orange3]! "
                           "Please add the following line to your configuration file:")
            logger.empty_line()
            logger.raw(os.linesep.join(configs), level=logging.WARNING)
            logger.empty_line()
            logger.empty_line()
            logger.info("Press enter to continue or Ctrl+C to cancel the operation")
            input()

    def __disableVPN(self) -> bool:
        """Remove a VPN profile for container startup (Only for interactive config)"""
        if self.__vpn_path:
            logger.verbose('Removing VPN configuration')
            self.__vpn_path = None
            self.__removeCapability("NET_ADMIN")
            self.__removeSysctl("net.ipv6.conf.all.disable_ipv6")
            self.removeDevice("/dev/net/tun")
            # Try to remove each possible volume
            self.removeVolume(container_path="/.exegol/vpn/auth/creds.txt")
            self.removeVolume(container_path="/.exegol/vpn/config/client.ovpn")
            self.removeVolume(container_path="/.exegol/vpn/config")
            self.__restoreEntrypoint()
            return True
        return False

    def disableDefaultWorkspace(self):
        """Allows you to disable the default workspace volume"""
        # If a custom workspace is not define, disable workspace
        if self.__workspace_custom_path is None:
            self.__disable_workspace = True

    def prepareShare(self, share_name: str):
        """Add workspace share before container creation"""
        for mount in self.__mounts:
            if mount.get('Target') == '/workspace':
                # Volume is already prepared
                return
        if self.__workspace_custom_path is not None:
            self.addVolume(self.__workspace_custom_path, '/workspace', enable_sticky_group=True)
        elif self.__disable_workspace:
            # Skip default volume workspace if disabled
            return
        else:
            # Add dedicated private workspace bind volume
            volume_path = str(UserConfig().private_volume_path.joinpath(share_name))
            self.addVolume(volume_path, '/workspace', enable_sticky_group=True)

    def rollback_preparation(self, share_name: str):
        """Undo preparation in case of container creation failure"""
        if self.__workspace_custom_path is None and not self.__disable_workspace:
            # Remove dedicated workspace volume
            logger.info("Rollback: removing dedicated workspace directory")
            directory_path = UserConfig().private_volume_path.joinpath(share_name)
            if directory_path.is_dir():
                directory_path.rmdir()

    def setNetworkMode(self, host_mode: Optional[bool]):
        """Set container's network mode, true for host, false for bridge"""
        if host_mode is None:
            host_mode = True
        if len(self.__ports) > 0 and host_mode:
            logger.warning("Host mode cannot be set with NAT ports configured. Disabling the shared network mode.")
            host_mode = False
        if EnvInfo.isDockerDesktop() and host_mode:
            logger.warning("Docker desktop (Windows & macOS) does not support sharing of host network interfaces.")
            logger.verbose("Official doc: https://docs.docker.com/network/host/")
            logger.info("To share network ports between the host and exegol, use the [bright_blue]--port[/bright_blue] parameter.")
            host_mode = False
        self.__network_host = host_mode

    def setContainerCommand(self, entrypoint_function: str, *parameters: str):
        """Set the entrypoint command of the container. This command is executed at each startup.
        This parameter is applied to the container at creation."""
        self.__container_command = [entrypoint_function] + list(parameters)

    def setLegacyContainerCommand(self, cmd: str):
        """Set the entrypoint command of the container. This command is executed at each startup.
        This parameter is applied to the container at creation.
        This method is legacy, before the entrypoint exist (support images before 3.x.x)."""
        self.__container_command_legacy = cmd

    def __restoreEntrypoint(self):
        """Restore container's entrypoint to its default configuration"""
        self.__container_command_legacy = None
        self.__container_command = self.__default_cmd

    def addCapability(self, cap_string: str):
        """Add a linux capability to the container"""
        if cap_string in self.__capabilities:
            logger.warning("Capability already setup. Skipping.")
            return
        self.__capabilities.append(cap_string)

    def __removeCapability(self, cap_string: str):
        """Remove a linux capability from the container's config"""
        try:
            self.__capabilities.remove(cap_string)
            return True
        except ValueError:
            # When the capability is not present
            return False

    def __addSysctl(self, sysctl_key: str, config: str):
        """Add a linux sysctl to the container"""
        if sysctl_key in self.__sysctls.keys():
            logger.warning(f"Sysctl {sysctl_key} already setup to '{self.__sysctls[sysctl_key]}'. Skipping.")
            return
        if self.__network_host:
            logger.warning(f"The sysctl container configuration is [red]not[/red] supported by docker in [blue]host[/blue] network mode.")
            logger.warning(f"Skipping the sysctl config: [magenta]{sysctl_key}[/magenta] = [orange3]{config}[/orange3].")
            logger.warning(f"If this configuration is mandatory in your situation, try to change it in sudo mode on your host.")
            return
        self.__sysctls[sysctl_key] = config

    def __removeSysctl(self, sysctl_key: str):
        """Remove a linux capability from the container's config"""
        try:
            self.__sysctls.pop(sysctl_key)
            return True
        except KeyError:
            # When the sysctl is not present
            return False

    def getNetworkMode(self) -> str:
        """Network mode, docker term getter"""
        return "host" if self.__network_host else "bridge"

    def getTextNetworkMode(self) -> str:
        """Network mode, text getter"""
        network_mode = "host" if self.__network_host else "bridge"
        if self.__vpn_path:
            network_mode += " with VPN"
        return network_mode

    def getPrivileged(self) -> bool:
        """Privileged getter"""
        return self.__privileged

    def getCapabilities(self) -> List[str]:
        """Capabilities getter"""
        return self.__capabilities

    def getSysctls(self) -> Dict[str, str]:
        """Sysctl custom rules getter"""
        return self.__sysctls

    def getWorkingDir(self) -> str:
        """Get default container's default working directory path"""
        return "/" if self.__disable_workspace else "/workspace"

    def getEntrypointCommand(self, image_entrypoint: Optional[Union[str, List[str]]]) -> Tuple[Optional[List[str]], Union[List[str], str]]:
        """Get container entrypoint/command arguments.
        This method support legacy configuration."""
        if image_entrypoint is None:
            # Legacy mode
            if self.__container_command_legacy is None:
                return [self.__default_entrypoint_legacy], []
            return None, self.__container_command_legacy
        else:
            return self.__container_entrypoint, self.__container_command

    def getShellCommand(self) -> str:
        """Get container command for opening a new shell"""
        # If shell logging was enabled at container creation, it'll always be enabled for every shell.
        # If not, it can be activated per shell basis
        if self.__shell_logging or ParametersManager().log:
            if self.__start_delegate_mode:
                # Use a start.sh script to handle the feature with the tools and feature corresponding to the image version
                # Start shell_logging feature using the user's specified method with the configured default shell w/ or w/o compression at the end
                return f"/.exegol/start.sh shell_logging {ParametersManager().log_method} {ParametersManager().shell} {UserConfig().shell_logging_compress ^ ParametersManager().log_compress}"
            else:
                # Legacy command support
                if ParametersManager().log_method != "script":
                    logger.warning("Your image version does not allow customization of the shell logging method. Using legacy script method.")
                compression_cmd = ''
                if UserConfig().shell_logging_compress ^ ParametersManager().log_compress:
                    compression_cmd = 'echo "Compressing logs, please wait..."; gzip $filelog; '
                return f"bash -c 'umask 007; mkdir -p /workspace/logs/; filelog=/workspace/logs/$(date +%d-%m-%Y_%H-%M-%S)_shell.log; script -qefac {ParametersManager().shell} $filelog; {compression_cmd}exit'"
        return ParametersManager().shell

    def getHostWorkspacePath(self) -> str:
        """Get private volume path (None if not set)"""
        if self.__workspace_custom_path:
            return FsUtils.resolvStrPath(self.__workspace_custom_path)
        elif self.__workspace_dedicated_path:
            return FsUtils.resolvStrPath(self.__workspace_dedicated_path)
        return "not found :("

    def getPrivateVolumePath(self) -> str:
        """Get private volume path (None if not set)"""
        return FsUtils.resolvStrPath(self.__workspace_dedicated_path)

    def isMyResourcesEnable(self) -> bool:
        """Return if the feature 'my-resources' is enabled in this container config"""
        return self.__my_resources

    def getMyResourcesPath(self) -> str:
        """Return if the feature 'exegol resources' is enabled in this container config"""
        return self.__my_resources_path

    def isExegolResourcesEnable(self) -> bool:
        """Return if the feature 'exegol resources' is enabled in this container config"""
        return self.__exegol_resources

    def isShellLoggingEnable(self) -> bool:
        """Return if the feature 'shell logging' is enabled in this container config"""
        return self.__shell_logging

    def isGUIEnable(self) -> bool:
        """Return if the feature 'GUI' is enabled in this container config"""
        return self.__enable_gui

    def isTimezoneShared(self) -> bool:
        """Return if the feature 'timezone' is enabled in this container config"""
        return self.__share_timezone

    def isWorkspaceCustom(self) -> bool:
        """Return if the workspace have a custom host volume"""
        return bool(self.__workspace_custom_path)

    def addVolume(self,
                  host_path: str,
                  container_path: str,
                  must_exist: bool = False,
                  read_only: bool = False,
                  enable_sticky_group: bool = False,
                  force_sticky_group: bool = False,
                  volume_type: str = 'bind'):
        """Add a volume to the container configuration.
        When the host path does not exist (neither file nor folder):
        if must_exist is set, an CancelOperation exception will be thrown.
        Otherwise, a folder will attempt to be created at the specified path.
        if set_sticky_group is set (on a Linux host), the permission setgid will be added to every folder on the volume."""
        # The creation of the directory is ignored when it is a path to the remote drive
        if volume_type == 'bind' and not host_path.startswith("\\\\"):
            path = Path(host_path)
            # TODO extend to docker desktop Windows
            if EnvInfo.isMacHost():
                # Add support for /etc
                path_match = str(path)
                if path_match.startswith("/etc/"):
                    if EnvInfo.isOrbstack():
                        raise CancelOperation(f"Orbstack doesn't support sharing /etc files with the container")
                    path_match = path_match.replace("/etc/", "/private/etc/")
                if EnvInfo.isDockerDesktop():
                    match = False
                    # Find a match
                    for resource in EnvInfo.getDockerDesktopResources():
                        if path_match.startswith(resource):
                            match = True
                            break
                    if not match:
                        logger.critical(f"Bind volume from {host_path} is not possible, Docker Desktop configuration is incorrect. "
                                        f"A parent directory must be shared in "
                                        f"[magenta]Docker Desktop > Preferences > Resources > File Sharing[/magenta].")
            # Choose to update fs directory perms if available and depending on user choice
            # if force_sticky_group is set, user choice is bypassed, fs will be updated.
            execute_update_fs = force_sticky_group or (enable_sticky_group and (UserConfig().auto_update_workspace_fs ^ ParametersManager().update_fs_perms))
            try:
                if not (path.is_file() or path.is_dir()):
                    if must_exist:
                        raise CancelOperation(f"{host_path} does not exist on your host.")
                    else:
                        # If the directory is created by exegol, bypass user preference and enable shared perms (if available)
                        execute_update_fs = force_sticky_group or enable_sticky_group
                        path.mkdir(parents=True, exist_ok=True)
            except PermissionError:
                logger.error("Unable to create the volume folder on the filesystem locally.")
                logger.critical(f"Insufficient permissions to create the folder: {host_path}")
            except FileExistsError:
                # The volume targets a file that already exists on the file system
                pass
            # Update FS don't work on Windows and only for directory
            if not EnvInfo.is_windows_shell and path.is_dir():
                if execute_update_fs:
                    # Apply perms update
                    FsUtils.setGidPermission(path)
                elif enable_sticky_group:
                    # If user choose not to update, print tips
                    logger.warning(f"The file sharing permissions between the container and the host will not be applied automatically by Exegol. ("
                                   f"{'Currently enabled by default according to the user config' if UserConfig().auto_update_workspace_fs else 'Use the --update-fs option to enable the feature'})")
        mount = Mount(container_path, host_path, read_only=read_only, type=volume_type)
        self.__mounts.append(mount)

    def addRawVolume(self, volume_string):
        """Add a volume to the container configuration from raw text input.
        Expected format is: /source/path:/target/mount:rw"""
        logger.debug(f"Parsing raw volume config: {volume_string}")
        parsing = re.match(r'^((\w:)?([\\/][\w .,:\-|()&;]*)+):(([\\/][\w .,\-|()&;]*)+)(:(ro|rw))?$',
                           volume_string)
        if parsing:
            host_path = parsing.group(1)
            container_path = parsing.group(4)
            mode = parsing.group(7)
            if mode is None or mode == "rw":
                readonly = False
            elif mode == "ro":
                readonly = True
            else:
                logger.error(f"Error on volume config, mode: {mode} not recognized.")
                readonly = False
            logger.debug(
                f"Adding a volume from '{host_path}' to '{container_path}' as {'readonly' if readonly else 'read/write'}")
            try:
                self.addVolume(host_path, container_path, readonly)
            except CancelOperation as e:
                logger.error(f"The following volume couldn't be created [magenta]{volume_string}[/magenta]. {e}")
                if not Confirm("Do you want to continue without this volume ?", False):
                    exit(0)
        else:
            logger.critical(f"Volume '{volume_string}' cannot be parsed. Exiting.")

    def removeVolume(self, host_path: Optional[str] = None, container_path: Optional[str] = None) -> bool:
        """Remove a volume from the container configuration (Only before container creation)"""
        if host_path is None and container_path is None:
            # This is a dev problem
            raise ValueError('At least one parameter must be set')
        for i in range(len(self.__mounts)):
            # For each Mount object compare the host_path if supplied or the container_path si supplied
            if host_path is not None and self.__mounts[i].get("Source") == host_path:
                # When the right object is found, remove it from the list
                self.__mounts.pop(i)
                return True
            if container_path is not None and self.__mounts[i].get("Target") == container_path:
                # When the right object is found, remove it from the list
                self.__mounts.pop(i)
                return True
        return False

    def getVolumes(self) -> List[Mount]:
        """Volume config getter"""
        return self.__mounts

    def __addDevice(self,
                    device_source: str,
                    device_dest: Optional[str] = None,
                    readonly: bool = False,
                    mknod: bool = False):
        """Add a device to the container configuration"""
        if device_dest is None:
            device_dest = device_source
        perm = 'r'
        if not readonly:
            perm += 'w'
        if mknod:
            perm += 'm'
        self.__devices.append(f"{device_source}:{device_dest}:{perm}")

    def addUserDevice(self, user_device_config: str):
        """Add a device from a user parameters"""
        if EnvInfo.isDockerDesktop():
            logger.warning("Docker desktop (Windows & macOS) does not support USB device passthrough.")
            logger.verbose("Official doc: https://docs.docker.com/desktop/faqs/#can-i-pass-through-a-usb-device-to-a-container")
            logger.critical("Device configuration cannot be applied, aborting operation.")
        self.__addDevice(user_device_config)

    def removeDevice(self, device_source: str) -> bool:
        """Remove a device from the container configuration (Only before container creation)"""
        for i in range(len(self.__devices)):
            # For each device, compare source device
            if self.__devices[i].split(':')[0] == device_source:
                # When found, remove it from the config list
                self.__devices.pop(i)
                return True
        return False

    def getDevices(self) -> List[str]:
        """Devices config getter"""
        return self.__devices

    def addEnv(self, key: str, value: str):
        """Add an environment variable to the container configuration"""
        self.__envs[key] = value

    def removeEnv(self, key: str) -> bool:
        """Remove an environment variable to the container configuration (Only before container creation)"""
        try:
            self.__envs.pop(key)
            return True
        except KeyError:
            # When the Key is not present in the dictionary
            return False

    def addRawEnv(self, env: str):
        """Parse and add an environment variable from raw user input"""
        key, value = self.__parseUserEnv(env)
        self.addEnv(key, value)

    def getEnvs(self) -> Dict[str, str]:
        """Envs config getter"""
        return self.__envs

    @classmethod
    def __parseUserEnv(cls, env: str) -> Tuple[str, str]:
        env_args = env.split('=')
        key = env_args[0]
        if len(env_args) < 2:
            value = os.getenv(env, '')
            if not value:
                logger.critical(f"Incorrect env syntax ({env}). Please use this format: KEY=value")
            else:
                logger.success(f"Using system value for env {env}.")
        else:
            value = '='.join(env_args[1:])
        return key, value

    def getShellEnvs(self) -> List[str]:
        """Overriding envs when opening a shell"""
        result = []
        if self.__enable_gui:
            current_display = GuiUtils.getDisplayEnv()
            # If the default DISPLAY environment in the container is not the same as the DISPLAY of the user's session,
            # the environment variable will be updated in the exegol shell.
            if current_display and self.__envs.get('DISPLAY', '') != current_display:
                # This case can happen when the container is created from a local desktop
                # but exegol can be launched from remote access via ssh with X11 forwarding
                # (Be careful, an .Xauthority file may be needed).
                result.append(f"DISPLAY={current_display}")
        # Overwrite env from user parameters
        user_envs = ParametersManager().envs
        if user_envs is not None:
            for env in user_envs:
                key, value = self.__parseUserEnv(env)
                logger.debug(f"Add env to current shell: {env}")
                result.append(f"{key}={value}")
        return result

    def addLabel(self, key: str, value: str):
        """Add a custom label to the container configuration"""
        self.__labels[key] = value

    def removeLabel(self, key: str) -> bool:
        """Remove a custom label from the container configuration (Only before container creation)"""
        try:
            self.__labels.pop(key)
            return True
        except KeyError:
            # When the Key is not present in the dictionary
            return False

    def getLabels(self) -> Dict[str, str]:
        """Labels config getter"""
        # Update metadata (from getter method) to the labels (on container creation)
        for label_name, refs in self.__label_metadata.items():
            data = getattr(self, refs[1])()
            if data is not None:
                self.addLabel(label_name, data)
        return self.__labels

    def getCreationDate(self) -> str:
        """Get container creation date.
        If the creation has not been set before, init as right now."""
        if self.creation_date is None:
            self.creation_date = datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
        return self.creation_date

    def getVpnName(self):
        """Get VPN Config name"""
        if self.__vpn_path is None:
            return "[bright_black]N/A[/bright_black]   "
        return f"[deep_sky_blue3]{self.__vpn_path.name}[/deep_sky_blue3]"

    def addPort(self,
                port_host: int,
                port_container: Union[int, str],
                protocol: str = 'tcp',
                host_ip: str = '0.0.0.0'):
        """Add port NAT config, only applicable on bridge network mode."""
        if self.__network_host:
            logger.warning("Port sharing is configured, disabling the host network mode.")
            self.setNetworkMode(False)
        if protocol.lower() not in ['tcp', 'udp', 'sctp']:
            raise ProtocolNotSupported(f"Unknown protocol '{protocol}'")
        logger.debug(f"Adding port {host_ip}:{port_host} -> {port_container}/{protocol}")
        self.__ports[f"{port_container}/{protocol}"] = (host_ip, port_host)

    def addRawPort(self, user_test_port: str):
        """Add port config from user input.
        Format must be [<host_ipv4>:]<host_port>[:<container_port>][:<protocol>]
        If host_ipv4 is not set, default to 0.0.0.0
        If container_port is not set, default is the same as host port
        If protocol is not set, default is 'tcp'"""
        match = re.search(r"^((\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}):)?(\d+):?(\d+)?:?(udp|tcp|sctp)?$", user_test_port)
        if match is None:
            logger.critical(f"Incorrect port syntax ({user_test_port}). Please use this format: [green][<host_ipv4>:]<host_port>[:<container_port>][:<protocol>][/green]")
            return
        host_ip = "0.0.0.0" if match.group(2) is None else match.group(2)
        protocol = "tcp" if match.group(5) is None else match.group(5)
        try:
            host_port = int(match.group(3))
            container_port = host_port if match.group(4) is None else int(match.group(4))
            if host_port > 65535 or container_port > 65535:
                raise ValueError
        except ValueError:
            logger.critical(f"The syntax for opening prot in NAT is incorrect. The ports must be numbers between 0 and 65535. ({match.group(3)}:{match.group(4)})")
            return
        self.addPort(host_port, container_port, protocol=protocol, host_ip=host_ip)

    def getPorts(self) -> Dict[str, Optional[Union[int, Tuple[str, int], List[int], List[Dict[str, Union[int, str]]]]]]:
        """Ports config getter"""
        return self.__ports

    def getTextFeatures(self, verbose: bool = False) -> str:
        """Text formatter for features configurations (Privileged, GUI, Network, Timezone, Shares)
        Print config only if they are different from their default config (or print everything in verbose mode)"""
        result = ""
        if verbose or self.__privileged:
            result += f"{getColor(not self.__privileged)[0]}Privileged: {'On :fire:' if self.__privileged else '[green]Off :heavy_check_mark:[/green]'}{getColor(not self.__privileged)[1]}{os.linesep}"
        if verbose or not self.__enable_gui:
            result += f"{getColor(self.__enable_gui)[0]}GUI: {boolFormatter(self.__enable_gui)}{getColor(self.__enable_gui)[1]}{os.linesep}"
        if verbose or not self.__network_host:
            result += f"[green]Network mode: [/green]{self.getTextNetworkMode()}{os.linesep}"
        if self.__vpn_path is not None:
            result += f"[green]VPN: [/green]{self.getVpnName()}{os.linesep}"
        if verbose or not self.__share_timezone:
            result += f"{getColor(self.__share_timezone)[0]}Share timezone: {boolFormatter(self.__share_timezone)}{getColor(self.__share_timezone)[1]}{os.linesep}"
        if verbose or not self.__exegol_resources:
            result += f"{getColor(self.__exegol_resources)[0]}Exegol resources: {boolFormatter(self.__exegol_resources)}{getColor(self.__exegol_resources)[1]}{os.linesep}"
        if verbose or not self.__my_resources:
            result += f"{getColor(self.__my_resources)[0]}My resources: {boolFormatter(self.__my_resources)}{getColor(self.__my_resources)[1]}{os.linesep}"
        if verbose or self.__shell_logging:
            result += f"{getColor(self.__shell_logging)[0]}Shell logging: {boolFormatter(self.__shell_logging)}{getColor(self.__shell_logging)[1]}{os.linesep}"
        result = result.strip()
        if not result:
            return "[i][bright_black]Default configuration[/bright_black][/i]"
        return result

    def getTextCreationDate(self) -> str:
        """Get the container creation date.
        If the creation date has not been supplied on the container, return empty string."""
        if self.creation_date is None:
            return ""
        return datetime.strptime(self.creation_date, "%Y-%m-%dT%H:%M:%SZ").strftime("%d/%m/%Y %H:%M")

    def getComment(self) -> Optional[str]:
        """Get the container comment. 
        If no comment has been supplied, returns None."""
        return self.comment

    def getTextMounts(self, verbose: bool = False) -> str:
        """Text formatter for Mounts configurations. The verbose mode does not exclude technical volumes."""
        result = ''
        for mount in self.__mounts:
            # Blacklist technical mount
            if not verbose and mount.get('Target') in ['/tmp/.X11-unix', '/opt/resources', '/etc/localtime',
                                                       '/etc/timezone', '/my-resources', '/opt/my-resources']:
                continue
            result += f"{mount.get('Source')} :right_arrow: {mount.get('Target')} {'(RO)' if mount.get('ReadOnly') else ''}{os.linesep}"
        return result

    def getTextDevices(self, verbose: bool = False) -> str:
        """Text formatter for Devices configuration. The verbose mode show full device configuration."""
        result = ''
        for device in self.__devices:
            if verbose:
                result += f"{device}{os.linesep}"
            else:
                src, dest = device.split(':')[:2]
                if src == dest:
                    result += f"{src}{os.linesep}"
                else:
                    result += f"{src}:right_arrow:{dest}{os.linesep}"
        return result

    def getTextEnvs(self, verbose: bool = False) -> str:
        """Text formatter for Envs configuration. The verbose mode does not exclude technical variables."""
        result = ''
        for k, v in self.__envs.items():
            # Blacklist technical variables, only shown in verbose
            if not verbose and k in list(self.__static_gui_envs.keys()) + ["DISPLAY", "PATH"]:
                continue
            result += f"{k}={v}{os.linesep}"
        return result

    def getTextPorts(self) -> str:
        """Text formatter for Ports configuration.
        Dict Port key = container port/protocol
        Dict Port Values:
          None = Random port
          int = open port ont he host
          tuple = (host_ip, port)
          list of int = open multiple host port
          list of dict = open one or more ports on host, key ('HostIp' / 'HostPort') and value ip or port"""
        result = ''
        for container_config, host_config in self.__ports.items():
            host_info = "Unknown"
            if host_config is None:
                host_info = "0.0.0.0:<Random port>"
            elif type(host_config) is int:
                host_info = f"0.0.0.0:{host_config}"
            elif type(host_config) is tuple:
                assert len(host_config) == 2
                host_info = f"{host_config[0]}:{host_config[1]}"
            elif type(host_config) is list:
                sub_info = []
                for sub_host_config in host_config:
                    if type(sub_host_config) is int:
                        sub_info.append(f"0.0.0.0:{sub_host_config}")
                    elif type(sub_host_config) is dict:
                        sub_port = sub_host_config.get('HostPort', '<Random port>')
                        if sub_port is None:
                            sub_port = "<Random port>"
                        sub_info.append(f"{sub_host_config.get('HostIp', '0.0.0.0')}:{sub_port}")
                if len(sub_info) > 0:
                    host_info = ", ".join(sub_info)
            else:
                logger.debug(f"Unknown port config: {type(host_config)}={host_config}")
            result += f"{host_info} :right_arrow: {container_config}{os.linesep}"
        return result

    def __str__(self):
        """Default object text formatter, debug only"""
        return f"Privileged: {self.__privileged}{os.linesep}" \
               f"Capabilities: {self.__capabilities}{os.linesep}" \
               f"Sysctls: {self.__sysctls}{os.linesep}" \
               f"X: {self.__enable_gui}{os.linesep}" \
               f"TTY: {self.tty}{os.linesep}" \
               f"Network host: {self.getNetworkMode()}{os.linesep}" \
               f"Ports: {self.__ports}{os.linesep}" \
               f"Share timezone: {self.__share_timezone}{os.linesep}" \
               f"Common resources: {self.__my_resources}{os.linesep}" \
               f"Envs ({len(self.__envs)}): {self.__envs}{os.linesep}" \
               f"Labels ({len(self.__labels)}): {self.__labels}{os.linesep}" \
               f"Shares ({len(self.__mounts)}): {self.__mounts}{os.linesep}" \
               f"Devices ({len(self.__devices)}): {self.__devices}{os.linesep}" \
               f"VPN: {self.getVpnName()}"

    def printConfig(self):
        """Log current object state, debug only"""
        logger.info(f"Current container config :{os.linesep}{self}")



from typing import cast, Union, Optional

from rich.console import Console
from rich.progress import Progress, Task, TaskID, ProgressColumn, GetTimeCallable

from exegol.utils.ExeLog import console as exelog_console


class ExegolProgress(Progress):
    """Addition of a practical function to Rich Progress"""

    def __init__(self, *columns: Union[str, ProgressColumn], console: Optional[Console] = None, auto_refresh: bool = True, refresh_per_second: float = 10, speed_estimate_period: float = 30.0,
                 transient: bool = False, redirect_stdout: bool = True, redirect_stderr: bool = True, get_time: Optional[GetTimeCallable] = None, disable: bool = False, expand: bool = False) -> None:
        if console is None:
            console = exelog_console
        super().__init__(*columns, console=console, auto_refresh=auto_refresh, refresh_per_second=refresh_per_second, speed_estimate_period=speed_estimate_period, transient=transient,
                         redirect_stdout=redirect_stdout, redirect_stderr=redirect_stderr, get_time=get_time, disable=disable, expand=expand)

    def getTask(self, task_id: TaskID) -> Task:
        """Return a specific task from task_id without error"""
        task = self._tasks.get(task_id)
        if task is None:
            # If task doesn't exist, raise IndexError exception
            raise IndexError
        return cast(Task, task)

    def __enter__(self) -> "ExegolProgress":
        super(ExegolProgress, self).__enter__()
        return self



#!/usr/bin/env python3
# -- coding: utf-8 --
# File name          : user-setup.py
# Author             : Skilo (@askilow - Alexis Marquois)
# Date created       : 07 march 2023
# Python Version     : 3.*

from setup import get_link, download_addon, read_manifest, install_addons, activate_addons
from R2Log import logger
from pathlib import Path
from glob import glob
from time import sleep
import re
import subprocess
import argparse

PATHNAME = "/root/.mozilla/firefox/**.Exegol/"
re_links = r'^https://addons\.mozilla\.org/fr/firefox/addon/[^/]+/?$'

def parse_args():
    arg_parser = argparse.ArgumentParser(description="Automatically installs addons from a list or folder containing .xpi files.")
    arg_parser.add_argument('-L', dest="addon_links", help="txt document containing addon link (ie: https://addons.mozilla.org/fr/firefox/addon/duckduckgo-for-firefox).")
    arg_parser.add_argument('-D', dest="addon_folder", help="Path to a folder containing .xpi files to install.")
    args = arg_parser.parse_args()
    return args

if __name__ == "__main__":

    args = parse_args()
    addon_links = args.addon_links
    addon_folder = args.addon_folder
    install_ok = False

    # Define a list containing all addons names and ids
    addon_list = []

    if addon_links is not None:
        # Read the list input by the user
        with open(addon_links, "r") as url_file:
            urls = url_file.read().splitlines()

        # Iterate through addons
        for url in urls:
            if re.findall(re_links, url):
                # Make a request to the URL
                link, addon_name = get_link(url)
                # Download the addon
                download_addon(link, addon_name)
                # Read manifest.json in the archive
                addon_id = read_manifest("/tmp/" + addon_name)
                install_addons(addon_name, addon_id, "/tmp/")
                logger.success(f"{addon_name} installed sucessfully\n")
                addon_list.append((addon_id, addon_name[0:-4], False))
                install_ok = True
        if install_ok:
            logger.success("All addons from the list were installed sucessfully\n")
        else:
            logger.error("No addons were found in the list %s.\n" % addon_links)

    if addon_folder is not None:
        if glob(addon_folder + "/*.xpi"):
            for addon_path in glob(addon_folder + "/*.xpi"):
                addon_name = addon_path.split("/")[-1]
                addon_id = read_manifest(addon_path)
                install_addons(addon_name, addon_id, addon_folder)
                logger.success(f"{addon_name} installed sucessfully\n")
                addon_list.append((addon_id, addon_name[0:-4], False))
                install_ok = True
            logger.success("All addons from the folder %s were installed sucessfully\n" % addon_folder)
        else:
            logger.error("No addons were found in the folder %s.\n" % addon_folder)

    if install_ok:
        # Run firefox to initialise profile
        logger.info("Initialising Firefox profile")
        try:
            p_firefox = subprocess.Popen(["firefox", "-P", "Exegol", "-headless"], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
            # Wait for firefox to be initialised
            while not addon_list[0][0].encode() in subprocess.check_output(["cat", glob("%s" % PATHNAME)[0] + "/extensions.json"]):
                sleep(0.5)
            p_firefox.kill()
            assert(Path(glob("%s" % PATHNAME)[0] + "/extensions.json").is_file())
            logger.success("Firefox profile initialised sucessfully\n")
        except:
            logger.error("Could not initialise Firefox profile")
            raise

        # Activate all addons
        activate_addons(addon_list)

        # Remove backup file interfering with addons activation
        logger.info("Removing backup file interfering with addons activation")
        try:
            Path(glob("%s" % PATHNAME)[0] + "/addonStartup.json.lz4").unlink()
            logger.success("Backup file successfully removed\n")
        except:
            logger.error("Could not remove the backup file")
            raise

        # Restart firefox to apply modifications
        logger.info("Restarting firefox to apply modifications")
        try:
            p_firefox = subprocess.Popen(["firefox", "-headless"], stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL)
            # Wait for modifications to be applied
            while not b'addonStartup.json.lz4' in subprocess.check_output(["ls", glob("%s" % PATHNAME)[0]]):
                sleep(0.5)
            p_firefox.kill()
            logger.success("Modifications successfully applied")
        except:
            logger.error("Could not restart firefox")
            raise
    else:
        logger.error("No addons were found.")



import os
import sys
from pathlib import Path
from typing import Optional, List

from git.exc import GitCommandError, RepositoryDirtyError
from rich.progress import TextColumn, BarColumn

from exegol.config.ConstantConfig import ConstantConfig
from exegol.config.EnvInfo import EnvInfo
from exegol.console.MetaGitProgress import MetaGitProgress, clone_update_progress, SubmoduleUpdateProgress
from exegol.console.cli.ParametersManager import ParametersManager
from exegol.utils.ExeLog import logger, console


# SDK Documentation : https://gitpython.readthedocs.io/en/stable/index.html

class GitUtils:
    """Utility class between exegol and the Git SDK"""

    def __init__(self,
                 path: Optional[Path] = None,
                 name: str = "wrapper",
                 subject: str = "source code",
                 skip_submodule_update: bool = False):
        """Init git local repository object / SDK"""
        if path is None:
            path = ConstantConfig.src_root_path_obj
        self.isAvailable = False
        self.__is_submodule = False
        self.__git_disable = False
        self.__repo_path = path
        self.__git_name: str = name
        self.__git_subject: str = subject
        abort_loading = False
        # Check if .git directory exist
        try:
            test_git_dir = self.__repo_path / '.git'
            if test_git_dir.is_file():
                logger.debug("Git submodule repository detected")
                self.__is_submodule = True
            elif not test_git_dir.is_dir():
                raise ReferenceError
            elif sys.platform == "win32":
                # Skip next platform specific code (temp fix for mypy static code analysis)
                pass
            elif not EnvInfo.is_windows_shell and test_git_dir.lstat().st_uid != os.getuid():
                raise PermissionError(test_git_dir.owner())
        except ReferenceError:
            if self.__git_name == "wrapper":
                logger.warning("Exegol has [red]not[/red] been installed via git clone. Skipping wrapper auto-update operation.")
                if ConstantConfig.pip_installed:
                    logger.info("If you have installed Exegol with pip, check for an update with the command "
                                "[green]pip3 install exegol --upgrade[/green]")
            abort_loading = True
        except PermissionError as e:
            logger.error(f"The repository {self.__git_name} has been cloned as [red]{e.args[0]}[/red].")
            logger.error("The current user does not have the necessary rights to perform the self-update operations.")
            logger.error("Please reinstall exegol (with git clone) without sudo.")
            abort_loading = True
        # locally import git in case git is not installed of the system
        try:
            from git import Repo, Remote, InvalidGitRepositoryError, FetchInfo
        except ModuleNotFoundError:
            self.__git_disable = True
            logger.warning("Git module is not installed. Python module 'GitPython' is missing, please install it with pip.")
            return
        except ImportError:
            self.__git_disable = True
            logger.error("Unable to find git tool locally. Skipping git operations.")
            return
        self.__gitRepo: Optional[Repo] = None
        self.__gitRemote: Optional[Remote] = None
        self.__fetchBranchInfo: Optional[FetchInfo] = None

        if abort_loading:
            return
        logger.debug(f"Loading git at {self.__repo_path}")
        try:
            self.__gitRepo = Repo(self.__repo_path)
            logger.debug(f"Repo path: {self.__gitRepo.git_dir}")
            self.__init_repo(skip_submodule_update)
        except InvalidGitRepositoryError as err:
            logger.verbose(err)
            logger.warning("Error while loading local git repository. Skipping all git operation.")

    def __init_repo(self, skip_submodule_update: bool = False):
        self.isAvailable = True
        assert self.__gitRepo is not None
        logger.debug("Git repository successfully loaded")
        if len(self.__gitRepo.remotes) > 0:
            self.__gitRemote = self.__gitRepo.remotes['origin']
        else:
            logger.warning("No remote git origin found on repository")
            logger.debug(self.__gitRepo.remotes)
        if not skip_submodule_update:
            self.__initSubmodules()

    def clone(self, repo_url: str, optimize_disk_space: bool = True) -> bool:
        if ParametersManager().offline_mode:
            logger.error("It's not possible to clone a repository in offline mode ...")
            return False
        if self.isAvailable:
            logger.warning(f"The {self.getName()} repo is already cloned.")
            return False
        # locally import git in case git is not installed of the system
        try:
            from git import Repo, Remote, InvalidGitRepositoryError, FetchInfo
        except ModuleNotFoundError:
            logger.debug("Git module is not installed.")
            return False
        except ImportError:
            logger.error(f"Unable to find git on your machine. The {self.getName()} repository cannot be cloned.")
            logger.warning("Please install git to support this feature.")
            return False
        custom_options = []
        if optimize_disk_space:
            custom_options.append('--depth=1')
        from git import GitCommandError
        try:
            with MetaGitProgress(TextColumn("{task.description}", justify="left"),
                                 BarColumn(bar_width=None),
                                 "•",
                                 "[progress.percentage]{task.percentage:>3.1f}%",
                                 "[bold]{task.completed}/{task.total}[/bold]") as progress:
                progress.add_task(f"[bold red]Cloning {self.getName()} git repository", start=True)
                self.__gitRepo = Repo.clone_from(repo_url, str(self.__repo_path), multi_options=custom_options, progress=clone_update_progress)
                progress.remove_task(progress.tasks[0].id)
            logger.success(f"The Git repository {self.getName()} was successfully cloned!")
        except GitCommandError as e:
            # GitPython user \n only
            error = GitUtils.formatStderr(e.stderr)
            logger.error(f"Unable to clone the git repository. {error}")
            return False
        self.__init_repo()
        return True

    def getCurrentBranch(self) -> Optional[str]:
        """Get current git branch name"""
        if not self.isAvailable:
            return None
        assert self.__gitRepo is not None
        try:
            return str(self.__gitRepo.active_branch)
        except TypeError:
            logger.debug("Git HEAD is detached, cant find the current branch.")
            return None
        except ValueError:
            logger.error(f"Unable to find current git branch in the {self.__git_name} repository. Check the path in the .git file from {self.__repo_path / '.git'}")
            return None

    def listBranch(self) -> List[str]:
        """Return a list of str of all remote git branch available"""
        assert self.isAvailable
        assert not ParametersManager().offline_mode
        result: List[str] = []
        if self.__gitRemote is None:
            return result
        for branch in self.__gitRemote.fetch():
            branch_parts = branch.name.split('/')
            if len(branch_parts) < 2:
                logger.warning(f"Branch name is not correct: {branch.name}")
                result.append(branch.name)
            else:
                result.append(branch_parts[1])
        return result

    def safeCheck(self) -> bool:
        """Check the status of the local git repository,
        if there is pending change it is not safe to apply some operations"""
        assert self.isAvailable
        if self.__gitRepo is None or self.__gitRemote is None:
            return False
        # Submodule changes must be ignored to update the submodules sources independently of the wrapper
        is_dirty = self.__gitRepo.is_dirty(submodules=False)
        if is_dirty:
            logger.warning("Local git have unsaved change. Skipping source update.")
        return not is_dirty

    def isUpToDate(self, branch: Optional[str] = None) -> bool:
        """Check if the local git repository is up-to-date.
        This method compare the last commit local and the ancestor."""
        assert self.isAvailable
        assert not ParametersManager().offline_mode
        if branch is None:
            branch = self.getCurrentBranch()
            if branch is None:
                logger.warning("No branch is currently attached to the git repository. The up-to-date status cannot be checked.")
                return False
        assert self.__gitRepo is not None
        assert self.__gitRemote is not None
        # Get last local commit
        current_commit = self.get_current_commit()
        # Get last remote commit
        if not self.__fetch_update(branch):
            return True

        assert self.__fetchBranchInfo is not None

        logger.debug(f"Fetch flags : {self.__fetchBranchInfo.flags}")
        logger.debug(f"Fetch note : {self.__fetchBranchInfo.note}")
        logger.debug(f"Fetch old commit : {self.__fetchBranchInfo.old_commit}")
        logger.debug(f"Fetch remote path : {self.__fetchBranchInfo.remote_ref_path}")
        from git import FetchInfo
        # Bit check to detect flags info
        if self.__fetchBranchInfo.flags & FetchInfo.HEAD_UPTODATE != 0:
            logger.debug("HEAD UP TO DATE flag detected")
        if self.__fetchBranchInfo.flags & FetchInfo.FAST_FORWARD != 0:
            logger.debug("FAST FORWARD flag detected")
        if self.__fetchBranchInfo.flags & FetchInfo.ERROR != 0:
            logger.debug("ERROR flag detected")
        if self.__fetchBranchInfo.flags & FetchInfo.FORCED_UPDATE != 0:
            logger.debug("FORCED_UPDATE flag detected")
        if self.__fetchBranchInfo.flags & FetchInfo.REJECTED != 0:
            logger.debug("REJECTED flag detected")
        if self.__fetchBranchInfo.flags & FetchInfo.NEW_TAG != 0:
            logger.debug("NEW TAG flag detected")

        remote_commit = self.get_latest_commit()
        assert remote_commit is not None
        # Check if remote_commit is an ancestor of the last local commit (check if there is local commit ahead)
        return self.__gitRepo.is_ancestor(remote_commit, current_commit)

    def __fetch_update(self, branch: Optional[str] = None) -> bool:
        """Fetch latest update from remote"""
        if self.__gitRemote is None:
            return False
        try:
            fetch_result = self.__gitRemote.fetch()
        except GitCommandError:
            logger.warning("Unable to fetch information from remote git repository, do you have internet ?")
            return False
        if branch is None:
            branch = self.getCurrentBranch()
        try:
            self.__fetchBranchInfo = fetch_result[f'{self.__gitRemote}/{branch}']
        except IndexError:
            logger.warning("The selected branch is local and cannot be updated.")
            return False
        return True

    def get_current_commit(self):
        """Fetch current commit id on the current branch."""
        assert self.isAvailable
        assert self.__gitRepo is not None
        branch = self.getCurrentBranch()
        if branch is None:
            branch = "master"
        # Get last local commit
        return self.__gitRepo.heads[branch].commit

    def get_latest_commit(self):
        """Fetch latest remote commit id on the current branch."""
        assert self.isAvailable
        assert not ParametersManager().offline_mode
        if self.__fetchBranchInfo is None:
            if not self.__fetch_update():
                logger.debug("The latest commit cannot be retrieved.")
                return None
        assert self.__fetchBranchInfo is not None
        return self.__fetchBranchInfo.commit

    def update(self) -> bool:
        """Update local git repository within current branch"""
        assert self.isAvailable
        assert not ParametersManager().offline_mode
        if not self.safeCheck():
            return False
        # Check if the git branch status is not detached
        if self.getCurrentBranch() is None:
            return False
        if self.isUpToDate():
            logger.info(f"Git branch [green]{self.getCurrentBranch()}[/green] is already up-to-date.")
            return False
        if self.__gitRemote is not None:
            logger.info(f"Using branch [green]{self.getCurrentBranch()}[/green] on {self.getName()} repository")
            with console.status(f"Updating git [green]{self.getName()}[/green]", spinner_style="blue"):
                self.__gitRemote.pull(refspec=self.getCurrentBranch())
            logger.success("Git successfully updated")
            return True
        return False

    def __initSubmodules(self):
        """Init (and update git object not source code) git sub repositories (only depth=1)"""
        if ParametersManager().offline_mode:
            logger.error("It's not possible to update any submodule in offline mode ...")
            return
        logger.verbose(f"Git {self.getName()} init submodules")
        # These modules are init / updated manually
        blacklist_heavy_modules = ["exegol-resources"]
        # Submodules dont have depth submodule limits
        depth_limit = not self.__is_submodule
        if self.__gitRepo is None:
            return
        with console.status(f"Initialization of git submodules", spinner_style="blue") as s:
            try:
                submodules = self.__gitRepo.iter_submodules()
            except ValueError:
                logger.error(f"Unable to find any git submodule from '{self.getName()}' repository. Check the path in the file {self.__repo_path / '.git'}")
                return
            for current_sub in submodules:
                # Submodule update are skipped if blacklisted or if the depth limit is set
                if current_sub.name in blacklist_heavy_modules or \
                        (depth_limit and ('/' in current_sub.name or '\\' in current_sub.name)):
                    continue
                s.update(status=f"Downloading git submodules [green]{current_sub.name}[/green]")
                from git.exc import GitCommandError
                try:
                    # TODO add TUI with progress
                    current_sub.update(recursive=True)
                except GitCommandError as e:
                    error = GitUtils.formatStderr(e.stderr)
                    logger.debug(f"Unable tu update git submodule {current_sub.name}: {e}")
                    if "unable to access" in error:
                        logger.error("You don't have internet to update git submodule. Skipping operation.")
                    else:
                        logger.error("Unable to update git submodule. Skipping operation.")
                        logger.error(error)
                except ValueError:
                    logger.error(f"Unable to update git submodule '{current_sub.name}'. Check the path in the file '{Path(current_sub.path) / '.git'}'")
                except RepositoryDirtyError as e:
                    logger.debug(e)
                    logger.error(f"Sub-repository {current_sub.name} have uncommitted local changes. Unable to automatically update this repository.")

    def submoduleSourceUpdate(self, name: str) -> bool:
        """Update source code from the 'name' git submodule"""
        assert not ParametersManager().offline_mode
        if not self.isAvailable:
            return False
        assert self.__gitRepo is not None
        try:
            submodule = self.__gitRepo.submodule(name)
        except ValueError:
            logger.debug(f"Git submodule '{name}' not found.")
            return False
        from git.exc import RepositoryDirtyError
        try:
            from git.exc import GitCommandError
            try:
                with MetaGitProgress(TextColumn("{task.description}", justify="left"),
                                     BarColumn(bar_width=None),
                                     "•",
                                     "[progress.percentage]{task.percentage:>3.1f}%",
                                     "[bold]{task.completed}/{task.total}[/bold]") as progress:
                    progress.add_task(f"[bold red]Downloading submodule [green]{name}[/green]", start=True)
                    submodule.update(to_latest_revision=True, progress=SubmoduleUpdateProgress())
                    progress.remove_task(progress.tasks[0].id)
            except GitCommandError as e:
                logger.debug(f"Unable tu update git submodule {name}: {e}")
                if "unable to access" in e.stderr:
                    logger.error("You don't have internet to update git submodule. Skipping operation.")
                else:
                    logger.error("Unable to update git submodule. Skipping operation.")
                    logger.error(e.stderr)
                return False
            logger.success(f"Submodule [green]{name}[/green] successfully updated.")
            return True
        except RepositoryDirtyError:
            logger.warning(f"Submodule {name} cannot be updated automatically as long as there are local modifications.")
            logger.error("Aborting git submodule update.")
        logger.empty_line()
        return False

    def checkout(self, branch: str) -> bool:
        """Change local git branch"""
        assert self.isAvailable
        if not self.safeCheck():
            return False
        if branch == self.getCurrentBranch():
            logger.warning(f"Branch '{branch}' is already the current branch")
            return False
        assert self.__gitRepo is not None
        from git.exc import GitCommandError
        try:
            # If git local branch didn't exist, change HEAD to the origin branch and create a new local branch
            if branch not in self.__gitRepo.heads:
                self.__gitRepo.references['origin/' + branch].checkout()
                self.__gitRepo.create_head(branch)
            self.__gitRepo.heads[branch].checkout()
        except GitCommandError as e:
            logger.error("Unable to checkout to the selected branch. Skipping operation.")
            logger.debug(e)
            return False
        except IndexError as e:
            logger.error("Unable to find the selected branch. Skipping operation.")
            logger.debug(e)
            return False
        logger.success(f"Git successfully checkout to '{branch}'")
        return True

    def getTextStatus(self) -> str:
        """Get text status from git object for rich print."""
        if self.isAvailable:
            from git.exc import GitCommandError
            try:
                if self.isUpToDate():
                    result = "[green]Up to date[/green]"
                else:
                    result = "[orange3]Update available[/orange3]"
            except GitCommandError:
                # Offline error catch
                result = "[green]Installed[/green] [bright_black](offline)[/bright_black]"
        else:
            if self.__git_disable:
                result = "[red]Missing dependencies[/red]"
            elif self.__git_name == ["wrapper", "images"] and \
                    (ConstantConfig.pip_installed or not ConstantConfig.git_source_installation):
                result = "[bright_black]Auto-update not supported[/bright_black]"
            else:
                result = "[bright_black]Not installed[/bright_black]"
        return result

    def getName(self) -> str:
        """Git name getter"""
        return self.__git_name

    def getSubject(self) -> str:
        """Git subject getter"""
        return self.__git_subject

    def isSubModule(self) -> bool:
        """Git submodule status getter"""
        return self.__is_submodule

    @classmethod
    def formatStderr(cls, stderr):
        return stderr.replace('\n', '').replace('stderr:', '').strip().strip("'")

    def __repr__(self) -> str:
        """Developer debug object representation"""
        return f"GitUtils '{self.__git_name}': {'Active' if self.isAvailable else 'Disable'}"



from typing import List, Optional

from argcomplete.completers import EnvironCompleter, DirectoriesCompleter, FilesCompleter

from exegol.config.UserConfig import UserConfig
from exegol.console.cli.ExegolCompleter import ContainerCompleter, ImageCompleter, VoidCompleter
from exegol.console.cli.actions.Command import Option, GroupArg


class ContainerSelector:
    """Generic parameter class for container selection"""

    def __init__(self, groupArgs: List[GroupArg]):
        # Create container selector arguments
        self.containertag: Optional[Option] = Option("containertag",
                                                     metavar="CONTAINER",
                                                     nargs='?',
                                                     action="store",
                                                     help="Tag used to target an Exegol container",
                                                     completer=ContainerCompleter)

        # Create group parameter for container selection
        groupArgs.append(GroupArg({"arg": self.containertag, "required": False},
                                  title="[blue]Container selection options[/blue]"))


class ContainerMultiSelector:
    """Generic parameter class for container multi selection"""

    def __init__(self, groupArgs: List[GroupArg]):
        # Create container selector arguments
        self.multicontainertag = Option("multicontainertag",
                                        metavar="CONTAINER",
                                        nargs='*',
                                        action="store",
                                        help="Tag used to target one or more Exegol containers",
                                        completer=ContainerCompleter)

        # Create group parameter for container multi selection
        groupArgs.append(GroupArg({"arg": self.multicontainertag, "required": False},
                                  title="[blue]Containers selection options[/blue]"))


class ContainerStart:
    """Generic parameter class for container selection.
    This generic class is used by start, restart and exec actions"""

    def __init__(self, groupArgs: List[GroupArg]):
        # Create options on container start
        self.envs = Option("-e", "--env",
                           action="append",
                           default=[],
                           dest="envs",
                           help="And an environment variable on Exegol (format: --env KEY=value). The variables "
                                "configured during the creation of the container will be persistent in all shells. "
                                "If the container already exists, the variable will be present only in the current shell",
                           completer=EnvironCompleter)

        # Create group parameter for container options at start
        groupArgs.append(GroupArg({"arg": self.envs, "required": False},
                                  title="[blue]Container start options[/blue]"))


class ContainerSpawnShell(ContainerStart):
    """Generic parameter class to spawn a shell on an exegol container.
    This generic class is used by start and restart"""

    def __init__(self, groupArgs: List[GroupArg]):
        # Spawn container shell arguments
        self.shell = Option("-s", "--shell",
                            dest="shell",
                            action="store",
                            choices=UserConfig.start_shell_options,
                            default=UserConfig().default_start_shell,
                            help=f"Select a shell environment to launch at startup (Default: [blue]{UserConfig().default_start_shell}[/blue])")

        self.log = Option("-l", "--log",
                          dest="log",
                          action="store_true",
                          default=False,
                          help="Enable shell logging (commands and outputs) on exegol to /workspace/logs/ (default: [red]Disabled[/red])")
        self.log_method = Option("--log-method",
                                 dest="log_method",
                                 action="store",
                                 choices=UserConfig.shell_logging_method_options,
                                 default=UserConfig().shell_logging_method,
                                 help=f"Select a shell logging method used to record the session (default: [blue]{UserConfig().shell_logging_method}[/blue])")
        self.log_compress = Option("--log-compress",
                                   dest="log_compress",
                                   action="store_true",
                                   default=False,
                                   help=f"Enable or disable the automatic compression of log files at the end of the session (default: {'[green]Enabled[/green]' if UserConfig().shell_logging_compress else '[red]Disabled[/red]'})")

        # Group dedicated to shell logging feature
        groupArgs.append(GroupArg({"arg": self.log, "required": False},
                                  {"arg": self.log_method, "required": False},
                                  {"arg": self.log_compress, "required": False},
                                  title="[blue]Container creation Shell logging options[/blue]"))

        ContainerStart.__init__(self, groupArgs)

        # Create group parameter for container selection
        groupArgs.append(GroupArg({"arg": self.shell, "required": False},
                                  title="[bold cyan]Start[/bold cyan] [blue]specific options[/blue]"))


class ImageSelector:
    """Generic parameter class for image selection"""

    def __init__(self, groupArgs: List[GroupArg]):
        # Create image selector arguments
        self.imagetag: Optional[Option] = Option("imagetag",
                                                 metavar="IMAGE",
                                                 nargs='?',
                                                 action="store",
                                                 help="Tag used to target an Exegol image",
                                                 completer=ImageCompleter)

        # Create group parameter for image selection
        groupArgs.append(GroupArg({"arg": self.imagetag, "required": False},
                                  title="[blue]Image selection options[/blue]"))


class ImageMultiSelector:
    """Generic parameter class for image multi selection"""

    def __init__(self, groupArgs: List[GroupArg]):
        # Create image multi selector arguments
        self.multiimagetag = Option("multiimagetag",
                                    metavar="IMAGE",
                                    nargs='*',
                                    action="store",
                                    help="Tag used to target one or more Exegol images",
                                    completer=ImageCompleter)

        # Create group parameter for image multi selection
        groupArgs.append(GroupArg({"arg": self.multiimagetag, "required": False},
                                  title="[blue]Images selection options[/blue]"))


class ContainerCreation(ContainerSelector, ImageSelector):
    """Generic parameter class for container creation"""

    def __init__(self, groupArgs: List[GroupArg]):
        # Init parents : ContainerStart > ContainerSelector
        ContainerSelector.__init__(self, groupArgs)
        ImageSelector.__init__(self, groupArgs)

        self.X11 = Option("--disable-X11",
                          action="store_false",
                          default=True,
                          dest="X11",
                          help="Disable display sharing to run GUI-based applications (default: [green]Enabled[/green])")
        self.my_resources = Option("--disable-my-resources",
                                   action="store_false",
                                   default=True,
                                   dest="my_resources",
                                   help=f"Disable the mount of the my-resources (/opt/my-resources) from the host ({UserConfig().my_resources_path}) (default: [green]Enabled[/green])")
        self.exegol_resources = Option("--disable-exegol-resources",
                                       action="store_false",
                                       default=True,
                                       dest="exegol_resources",
                                       help=f"Disable the mount of the exegol resources (/opt/resources) from the host ({UserConfig().exegol_resources_path}) (default: [green]Enabled[/green])")
        self.host_network = Option("--disable-shared-network",
                                   action="store_false",
                                   default=True,
                                   dest="host_network",
                                   help="Disable the sharing of the host's network interfaces with exegol (default: [green]Enabled[/green])")
        self.share_timezone = Option("--disable-shared-timezones",
                                     action="store_false",
                                     default=True,
                                     dest="share_timezone",
                                     help="Disable the sharing of the host's time and timezone configuration with exegol (default: [green]Enabled[/green])")
        self.mount_current_dir = Option("-cwd", "--cwd-mount",
                                        dest="mount_current_dir",
                                        action="store_true",
                                        default=False,
                                        help="This option is a shortcut to set the /workspace folder to the user's current working directory")
        self.workspace_path = Option("-w", "--workspace",
                                     dest="workspace_path",
                                     action="store",
                                     help="The specified host folder will be linked to the /workspace folder in the container",
                                     completer=DirectoriesCompleter())
        self.update_fs_perms = Option("-fs", "--update-fs",
                                      action="store_true",
                                      default=False,
                                      dest="update_fs_perms",
                                      help=f"Modifies the permissions of folders and sub-folders shared in your workspace to access the files created within the container using your host user account. "
                                           f"(default: {'[green]Enabled[/green]' if UserConfig().auto_update_workspace_fs else '[red]Disabled[/red]'})")
        self.volumes = Option("-V", "--volume",
                              action="append",
                              default=[],
                              dest="volumes",
                              help="Share a new volume between host and exegol (format: --volume /path/on/host/:/path/in/container/)")
        self.ports = Option("-p", "--port",
                            action="append",
                            default=[],
                            dest="ports",
                            help="Share a network port between host and exegol (format: --port [<host_ipv4>:]<host_port>[:<container_port>][:<protocol>]. This configuration will disable the shared network with the host.",
                            completer=VoidCompleter)
        self.hostname = Option("--hostname",
                               dest="hostname",
                               default=None,
                               action="store",
                               help="Set a custom hostname to the exegol container (default: exegol-<name>)",
                               completer=VoidCompleter)
        self.capabilities = Option("--cap",
                                   dest="capabilities",
                                   metavar='CAP',  # Do not display available choices
                                   action="append",
                                   default=[],
                                   choices={"NET_ADMIN", "NET_BROADCAST", "SYS_MODULE", "SYS_PTRACE", "SYS_RAWIO",
                                            "SYS_ADMIN", "LINUX_IMMUTABLE", "MAC_ADMIN", "SYSLOG"},
                                   help="[orange3](dangerous)[/orange3] Capabilities allow to add [orange3]specific[/orange3] privileges to the container "
                                        "(e.g. need to mount volumes, perform low-level operations on the network, etc).")
        self.privileged = Option("--privileged",
                                 dest="privileged",
                                 action="store_true",
                                 default=False,
                                 help="[orange3](dangerous)[/orange3] Give [red]ALL[/red] admin privileges to the container when it is created "
                                      "(if the need is specifically identified, consider adding capabilities instead).")
        self.devices = Option("-d", "--device",
                              dest="devices",
                              default=[],
                              action="append",
                              help="Add host [default not bold]device(s)[/default not bold] at the container creation (example: -d /dev/ttyACM0 -d /dev/bus/usb/)")

        self.vpn = Option("--vpn",
                          dest="vpn",
                          default=None,
                          action="store",
                          help="Setup an OpenVPN connection at the container creation (example: --vpn /home/user/vpn/conf.ovpn)",
                          completer=FilesCompleter(["ovpn"], directories=True))
        self.vpn_auth = Option("--vpn-auth",
                               dest="vpn_auth",
                               default=None,
                               action="store",
                               help="Enter the credentials with a file (first line: username, second line: password) to establish the VPN connection automatically (example: --vpn-auth /home/user/vpn/auth.txt)")

        self.comment = Option("--comment",
                              dest="comment",
                              action="store",
                              help="The specified comment will be added to the container info",
                              completer=VoidCompleter)

        groupArgs.append(GroupArg({"arg": self.workspace_path, "required": False},
                                  {"arg": self.mount_current_dir, "required": False},
                                  {"arg": self.update_fs_perms, "required": False},
                                  {"arg": self.volumes, "required": False},
                                  {"arg": self.ports, "required": False},
                                  {"arg": self.hostname, "required": False},
                                  {"arg": self.capabilities, "required": False},
                                  {"arg": self.privileged, "required": False},
                                  {"arg": self.devices, "required": False},
                                  {"arg": self.X11, "required": False},
                                  {"arg": self.my_resources, "required": False},
                                  {"arg": self.exegol_resources, "required": False},
                                  {"arg": self.host_network, "required": False},
                                  {"arg": self.share_timezone, "required": False},
                                  {"arg": self.comment, "required": False},
                                  title="[blue]Container creation options[/blue]"))

        groupArgs.append(GroupArg({"arg": self.vpn, "required": False},
                                  {"arg": self.vpn_auth, "required": False},
                                  title="[blue]Container creation VPN options[/blue]"))



# Generic singleton class
from typing import Dict


class MetaSingleton(type):
    """Metaclass to create a singleton class"""
    __instances: Dict[type, object] = {}

    def __call__(cls, *args, **kwargs) -> object:
        """Redirects each call to the current class to the corresponding single instance"""
        if cls not in MetaSingleton.__instances:
            # If the instance does not already exist, it is created
            MetaSingleton.__instances[cls] = super(MetaSingleton, cls).__call__(*args, **kwargs)
        # Return the desired object
        return MetaSingleton.__instances[cls]



from argparse import Namespace
from typing import Tuple

from exegol.config.DataCache import DataCache
from exegol.manager.UpdateManager import UpdateManager
from exegol.utils.DockerUtils import DockerUtils


def ContainerCompleter(prefix: str, parsed_args: Namespace, **kwargs) -> Tuple[str, ...]:
    """Function to dynamically load a container list for CLI autocompletion purpose"""
    data = [c.name for c in DockerUtils.listContainers()]
    for obj in data:
        # filter data if needed
        if prefix and not obj.lower().startswith(prefix.lower()):
            data.remove(obj)
    return tuple(data)


def ImageCompleter(prefix: str, parsed_args: Namespace, **kwargs) -> Tuple[str, ...]:
    """Function to dynamically load an image list for CLI autocompletion purpose"""
    # Skip image completer when container hasn't been selected first (because parameters are all optional, parameters order is not working)
    if parsed_args is not None and str(parsed_args.action) == "start" and parsed_args.containertag is None:
        return ()
    try:
        if parsed_args is not None and str(parsed_args.action) == "install":
            data = [img_cache.name for img_cache in DataCache().get_images_data().data if img_cache.source == "remote"]
        else:
            data = [img_cache.name for img_cache in DataCache().get_images_data().data]
    except Exception as e:
        data = []
    if len(data) == 0:
        # Fallback with default data if the cache is not initialized yet
        data = ["full", "nightly", "ad", "web", "light", "osint"]
    for obj in data:
        # filter data if needed
        if prefix and not obj.lower().startswith(prefix.lower()):
            data.remove(obj)
    return tuple(data)


def HybridContainerImageCompleter(prefix: str, parsed_args: Namespace, **kwargs) -> Tuple[str, ...]:
    """Hybrid completer for auto-complet. The selector on exec action is hybrid between image and container depending on the mode (tmp or not).
    This completer will supply the adequate data."""
    # "exec" parameter is filled first before the selector argument
    # If "selector" is null but the selector parameter is set in the first exec slot, no longer need to supply completer options
    if parsed_args.selector is None and parsed_args.exec is not None and len(parsed_args.exec) > 0:
        return ()
    # In "tmp" mode, the user must choose an image, otherwise it's a container
    if parsed_args.tmp:
        return ImageCompleter(prefix, parsed_args, **kwargs)
    else:
        return ContainerCompleter(prefix, parsed_args, **kwargs)


def BuildProfileCompleter(prefix: str, parsed_args: Namespace, **kwargs) -> Tuple[str, ...]:
    """Completer function for build profile parameter. The completer must be trigger only when an image name have already been chosen."""
    # The build profile completer must be trigger only when an image name have been set by user
    if parsed_args is not None and parsed_args.imagetag is None:
        return ()
    data = list(UpdateManager.listBuildProfiles().keys())
    for obj in data:
        if prefix and not obj.lower().startswith(prefix.lower()):
            data.remove(obj)
    return tuple(data)


def VoidCompleter(**kwargs) -> Tuple:
    """No option to auto-complet"""
    return ()



import io
import os
import shutil
import subprocess
import sys
import time
from pathlib import Path
from typing import Optional

from exegol.console.ExegolPrompt import Confirm
from exegol.exceptions.ExegolExceptions import CancelOperation
from exegol.config.EnvInfo import EnvInfo
from exegol.utils.ExeLog import logger, console


class GuiUtils:
    """This utility class allows determining if the current system supports the GUI
    from the information of the system."""

    __distro_name = ""
    default_x11_path = "/tmp/.X11-unix"

    @classmethod
    def isGuiAvailable(cls) -> bool:
        """
        Check if the host OS can support GUI application with X11 sharing
        :return: bool
        """
        # GUI was not supported on Windows before WSLg
        if EnvInfo.isWindowsHost():
            return cls.__windowsGuiChecks()
        elif EnvInfo.isMacHost():
            return cls.__macGuiChecks()
        # Linux default is True
        return True

    @classmethod
    def getX11SocketPath(cls) -> Optional[str]:
        """
        Get the host path of the X11 socket
        :return:
        """
        if cls.__distro_name:
            # Distro name can only be set if the current host OS is Windows
            return f"\\\\wsl.localhost\\{cls.__distro_name}\\mnt\\wslg\\.X11-unix"
        elif EnvInfo.isWindowsHost():
            if EnvInfo.current_platform == "WSL":
                # Mount point from a WSL shell context
                return f"/mnt/wslg/.X11-unix"
            else:
                # From a Windows context, a WSL distro should have been supply during GUI checks
                logger.debug(f"No WSL distro have been previously found: '{cls.__distro_name}'")
                raise CancelOperation("Exegol tried to create a container with GUI support on a Windows host "
                                      "without having performed the availability tests before.")
        elif EnvInfo.isMacHost():
            # Docker desktop don't support UNIX socket through volume, we are using XQuartz over the network until then
            return None
        # Other distributions (Linux / Mac) have the default socket path
        return cls.default_x11_path

    @classmethod
    def getDisplayEnv(cls) -> str:
        """
        Get the current DISPLAY env to access X11 socket
        :return:
        """
        if EnvInfo.isMacHost():
            # xquartz Mac mode
            return "host.docker.internal:0"

        # Add ENV check is case of user don't have it, which will mess up GUI if fallback does not work
        # @see https://github.com/ThePorgs/Exegol/issues/148
        if os.getenv("DISPLAY") is None:
            logger.warning("The DISPLAY environment variable is not set on your host. This can prevent GUI apps to start")

        # DISPLAY var is fetch from the current user environment. If it doesn't exist, using ':0'.
        return os.getenv('DISPLAY', ":0")

    # # # # # # Mac specific methods # # # # # #

    @classmethod
    def __macGuiChecks(cls) -> bool:
        """
        Procedure to check if the Mac host supports GUI with docker through XQuartz
        :return: bool
        """
        if not cls.__isXQuartzInstalled():
            logger.warning("Display sharing is [orange3]not supported[/orange3] on your mac without XQuartz installed. "
                           "You need to manually install [turquoise2]XQuartz[/turquoise2] and check the configuration 'Allow connections from network clients'.")
            return False
        logger.debug("XQuartz detected.")
        if not cls.__xquartzAllowNetworkClients():
            # Notify user to change configuration
            logger.error("XQuartz does not allow network connections. "
                         "You need to manually change the configuration to 'Allow connections from network clients'")
            # Add sys.platform check to exclude windows env (fix for mypy static code analysis)
            if sys.platform != "win32" and os.getuid() == 0:
                logger.warning("You are running exegol as [red]root[/red]! The root user cannot check in the user context whether XQuartz is properly configured or not.")
            return False

        # Check if XQuartz is started, check is dir exist and if there is at least one socket
        if not cls.__isXQuartzRunning():
            if not cls.__startXQuartz():
                logger.warning("Unable to start XQuartz service.")
                return False

        # The /tmp config is not necessary until you can use the unix socket with docker-desktop volume
        # Check if Docker Desktop is configured with /tmp in Docker Desktop > Preferences > Resources > File Sharing
        #if EnvInfo.isDockerDesktop() and not cls.__checkDockerDesktopResourcesConfig():
        #    logger.warning("Display sharing not possible, Docker Desktop configuration is incorrect. Please add /tmp in "
        #                   "[magenta]Docker Desktop > Preferences > Resources > File Sharing[/magenta]")
        #    return False
        return True

    @staticmethod
    def __checkDockerDesktopResourcesConfig() -> bool:
        """
            Check if Docker Desktop for macOS is configured correctly, allowing
            /tmp to be bind mounted into Docker containers, which is needed to
             mount /tmp/.X11-unix for display sharing.
             Return True if the configuration is correct and /tmp is part of the whitelisted resources
        """
        docker_config = EnvInfo.getDockerDesktopResources()
        logger.debug(f"Docker Desktop configuration filesharingDirectories: {docker_config}")
        return '/tmp' in docker_config

    @staticmethod
    def __isXQuartzInstalled() -> bool:
        return 'xquartz' in os.getenv('DISPLAY', "").lower()

    @staticmethod
    def __xquartzAllowNetworkClients() -> bool:
        # defaults read org.xquartz.X11.plist nolisten_tcp
        conf_check = subprocess.run(["defaults", "read", "org.xquartz.X11.plist", "nolisten_tcp"],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE)
        logger.debug(f"XQuartz nolisten_tcp config: '{conf_check.stdout.strip().decode('utf-8')}'")
        return conf_check.stdout.strip() == b'0'

    @classmethod
    def __isXQuartzRunning(cls) -> bool:
        """
        Check if xquartz service is up by testing sockets
        """
        socket_path = Path(cls.default_x11_path)
        socket_x11_found = False
        if socket_path.is_dir():
            for file in socket_path.glob("*"):
                if file.is_socket():
                    socket_x11_found = True
                    break
        return socket_x11_found

    @staticmethod
    def __startXQuartz() -> bool:
        xhost_path = shutil.which("xhost")
        if not xhost_path:
            logger.error("xhost command not found, check your XQuartz installation")
            return False
        # Starting xquartz
        logger.debug("Starting XQuartz using xhost command")
        with console.status(f"Starting [green]XQuartz[/green]...", spinner_style="blue"):
            run_xhost = subprocess.run([xhost_path], shell=True,
                                       stdout=subprocess.DEVNULL,
                                       stderr=subprocess.DEVNULL)
        return run_xhost.returncode == 0

    # # # # # # Windows specific methods # # # # # #

    @classmethod
    def __windowsGuiChecks(cls) -> bool:
        """
        Procedure to check if the Windows host supports GUI with docker through WSLg
        :return: bool
        """
        logger.debug("Testing WSLg availability")
        # WSL + WSLg must be available on the Windows host for the GUI to work
        if not cls.__wsl_available():
            logger.error("WSL is [orange3]not available[/orange3] on your system. GUI is not supported.")
            return False
        # Only WSL2 support WSLg
        if EnvInfo.getDockerEngine() != EnvInfo.DockerEngine.WLS2:
            logger.error("Docker must be run with [orange3]WSL2[/orange3] engine in order to support GUI applications.")
            return False
        logger.debug("WSL is [green]available[/green] and docker is using WSL2")
        # X11 GUI socket can only be shared from a WSL (to find WSLg mount point)
        if EnvInfo.current_platform != "WSL":
            logger.debug("Exegol is running from a Windows context (e.g. Powershell), a WSL instance must be found to share WSLg X11 socket")
            cls.__distro_name = cls.__find_wsl_distro()
            logger.debug(f"Set WSL Distro as: '{cls.__distro_name}'")
            # If no WSL is found, propose to continue without GUI
            if not cls.__distro_name and not Confirm(
                    "Do you want to continue [orange3]without[/orange3] GUI support ?", default=True):
                raise KeyboardInterrupt
        else:
            logger.debug("Using current WSL context for X11 socket sharing")
        if cls.__wslg_installed():
            logger.debug("WSLg seems to be installed.")
            return True
        elif cls.__wslg_eligible():
            logger.info("[green]WSLg[/green] is available on your system but [orange3]not installed[/orange3].")
            logger.info("Make sure, [green]WSLg[/green] is installed on your Windows by running "
                        "'wsl --update' as [orange3]admin[/orange3].")
            return True
        logger.debug("WSLg is [orange3]not available[/orange3]")
        logger.warning("Display sharing is [orange3]not supported[/orange3] on your version of Windows. "
                       "You need to upgrade to [turquoise2]Windows 11[/turquoise2].")
        return False

    @staticmethod
    def __wsl_test(path, name: Optional[str] = "docker-desktop") -> bool:
        """
        Check presence of a file in the WSL docker-desktop image.
        the targeted WSL image can be changed with 'name' parameter.
        If name is None, the default WSL image will be use.
        """
        if EnvInfo.isWindowsHost():
            wsl = shutil.which("wsl.exe")
            if not wsl:
                return False
            if name is None:
                ret = subprocess.run(["wsl.exe", "test", "-f", path])
            else:
                ret = subprocess.run(["wsl.exe", "-d", name, "test", "-f", path])
            return ret.returncode == 0
        return False

    @classmethod
    def __check_wsl_docker_integration(cls, distrib_name) -> bool:
        """
        Check the presence of the docker binary in the supplied WSL distribution.
        This test allows checking if docker integration is enabled.
        """
        return cls.__wsl_test("/usr/bin/docker", distrib_name)

    @classmethod
    def __wsl_available(cls) -> bool:
        """
        heuristic to detect if Windows Subsystem for Linux is available.

        Uses presence of /etc/os-release in the WSL image to say Linux is there.
        This is a de facto file standard across Linux distros.

        Tests the existence of WSL by searching in the default WSL first.
        However, if the default wsl is 'docker-desktop-data', the result will be false, so you have to test with docker-desktop.
        """
        if EnvInfo.isWindowsHost():
            wsl = shutil.which("wsl.exe")
            if not wsl:
                return False
            ret = subprocess.Popen(["wsl.exe", "--status"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            ret.wait()
            if ret.returncode == 0:
                return True
        logger.debug("WSL status command failed.. Trying a fallback check method.")
        return cls.__wsl_test("/etc/os-release", name=None) or cls.__wsl_test("/etc/os-release")

    @classmethod
    def __wslg_installed(cls) -> bool:
        """
        Check if WSLg is installed and deploy inside a WSL image by testing if the file wslg/versions.txt exist.
        :return: bool
        """
        if EnvInfo.current_platform == "WSL":
            if Path("/mnt/host/wslg/versions.txt").is_file():
                return True
        else:
            if cls.__wsl_test("/mnt/host/wslg/versions.txt", name=cls.__distro_name):
                return True
        logger.debug("WSLg check failed.. Trying a fallback check method.")
        return cls.__wsl_test("/mnt/host/wslg/versions.txt") or cls.__wsl_test("/mnt/wslg/versions.txt", name=None)

    @staticmethod
    def __wslg_eligible() -> bool:
        """
        Check if the current Windows version support WSLg
        :return:
        """
        try:
            os_version_raw, _, build_number_raw = EnvInfo.getWindowsRelease().split('.')[:3]
        except ValueError:
            logger.debug(f"Impossible to find the version of windows: '{EnvInfo.getWindowsRelease()}'")
            logger.error("Exegol can't know if your [orange3]version of Windows[/orange3] can support dockerized GUIs.")
            return False
        # Available from Windows 10 Build 21364
        # Available from Windows 11 Build 22000
        os_version = int(os_version_raw)
        build_number = int(build_number_raw)
        if os_version == 10 and build_number >= 21364:
            return True
        elif os_version > 10:
            return True
        return False

    @classmethod
    def __find_wsl_distro(cls) -> str:
        distro_name = ""
        # these distros cannot be used to load WSLg socket
        blacklisted_distro = ["docker-desktop", "docker-desktop-data"]
        ret = subprocess.Popen(["C:\Windows\system32\wsl.exe", "-l"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        # Wait for WSL process to end
        ret.wait()
        if ret.returncode == 0:
            skip_header = True
            # parse distribs
            logger.debug("Found WSL distribution:")
            assert ret.stdout is not None
            for line in io.TextIOWrapper(ret.stdout, encoding="utf-16le"):
                # Skip WSL text header
                if skip_header:
                    skip_header = False
                    continue
                # Remove newline
                line = line.strip()
                # Skip if line is empty
                if not line:
                    continue
                # Remove default text message
                name = line.split()[0]
                logger.debug(f"- {name}")
                # Skip blacklisted WSL
                if name not in blacklisted_distro:
                    eligible = True
                    # Test if the current WSL has docker integration activated
                    while not cls.__check_wsl_docker_integration(name):
                        eligible = False
                        logger.warning(
                            f"The '{name}' WSL distribution could be used to [green]enable the GUI[/green] on exegol but the docker integration is [orange3]not enabled[/orange3].")
                        if not Confirm(
                                f"Do you want to [red]manually[/red] enable docker integration for WSL '{name}'?",
                                default=True):
                            break
                        eligible = True
                    if eligible:
                        distro_name = name
                        break
            if distro_name:
                logger.verbose(f"Wsl '{distro_name}' distribution found, the WSLg service can be mounted in exegol.")
            else:
                logger.warning(
                    "No WSL distribution was found on your machine. At least one distribution must be available to allow Exegol to use WSLg.")
                if Confirm("Do you want Exegol to install one automatically (Ubuntu)?", default=True):
                    if cls.__create_default_wsl():
                        distro_name = "Ubuntu"
        else:
            assert ret.stderr is not None
            logger.error(
                f"Error while loading existing wsl distributions. {ret.stderr.read().decode('utf-16le')} (code: {ret.returncode})")
        return distro_name

    @classmethod
    def __create_default_wsl(cls) -> bool:
        logger.info("Creating Ubuntu WSL distribution. Please wait.")
        ret = subprocess.Popen(["C:\Windows\system32\wsl.exe", "--install", "-d", "Ubuntu"], stderr=subprocess.PIPE)
        ret.wait()
        logger.info("Please follow installation instructions on the new window.")
        if ret.returncode != 0:
            assert ret.stderr is not None
            logger.error(
                f"Error while install WSL Ubuntu: {ret.stderr.read().decode('utf-16le')} (code: {ret.returncode})")
            return False
        else:
            while not Confirm("Is the installation of Ubuntu [green]finished[/green]?", default=True):
                pass
            # Check if docker have default docker integration
            docker_settings = EnvInfo.getDockerDesktopSettings()
            if docker_settings is not None and docker_settings.get("enableIntegrationWithDefaultWslDistro", False):
                logger.verbose("Set WSL Ubuntu as default to automatically enable docker integration")
                # Set new WSL distribution as default to start it and enable docker integration
                ret = subprocess.Popen(["C:\Windows\system32\wsl.exe", "-s", "Ubuntu"], stderr=subprocess.PIPE)
                ret.wait()
                # Wait for the docker integration (10 try, 1 sec apart)
                with console.status("Waiting for the activation of the docker integration", spinner_style="blue"):
                    for _ in range(10):
                        if cls.__check_wsl_docker_integration("Ubuntu"):
                            break
                        time.sleep(1)
            while not cls.__check_wsl_docker_integration("Ubuntu"):
                logger.error("The newly created WSL could not get the docker integration automatically. "
                             "It has to be activated [red]manually[/red]")
                logger.info("Enable WSL Docker integration for the newly created WSL in: [magenta]Docker Desktop > Settings > Resources > WSL Integration[/magenta]")
                if not Confirm("Has the WSL Ubuntu docker integration been [red]manually[/red] activated?",
                               default=True):
                    return False
            logger.success("WSL 'Ubuntu' successfully created with docker integration")
            return True



from pathlib import Path
from typing import Optional, Union

from exegol.console.ExegolPrompt import Confirm
from exegol.console.cli.ParametersManager import ParametersManager
from exegol.exceptions.ExegolExceptions import CancelOperation
from exegol.config.ConstantConfig import ConstantConfig
from exegol.utils.ExeLog import logger
from exegol.utils.GitUtils import GitUtils
from exegol.utils.MetaSingleton import MetaSingleton
from exegol.config.UserConfig import UserConfig


class ExegolModules(metaclass=MetaSingleton):
    """Singleton class dedicated to the centralized management of the project modules"""

    def __init__(self):
        """Init project git modules to None until their first call"""
        # Git modules
        self.__git_wrapper: Optional[GitUtils] = None
        self.__git_source: Optional[GitUtils] = None
        self.__git_resources: Optional[GitUtils] = None

        # Git loading mode
        self.__wrapper_fast_loaded = False

    def getWrapperGit(self, fast_load: bool = False) -> GitUtils:
        """GitUtils local singleton getter.
        Set fast_load to True to disable submodule init/update."""
        # If the module have been previously fast loaded and must be reuse later in standard mode, it can be recreated
        if self.__git_wrapper is None or (not fast_load and self.__wrapper_fast_loaded):
            self.__wrapper_fast_loaded = fast_load
            self.__git_wrapper = GitUtils(skip_submodule_update=fast_load)
        return self.__git_wrapper

    def getSourceGit(self, fast_load: bool = False) -> GitUtils:
        """GitUtils source submodule singleton getter.
        Set fast_load to True to disable submodule init/update."""
        # Be sure that submodules are init first
        self.getWrapperGit()
        if self.__git_source is None:
            self.__git_source = GitUtils(ConstantConfig.src_root_path_obj / "exegol-docker-build", "images",
                                         skip_submodule_update=fast_load)
        return self.__git_source

    def getResourcesGit(self, fast_load: bool = False, skip_install: bool = False) -> GitUtils:
        """GitUtils resource repo/submodule singleton getter.
        Set fast_load to True to disable submodule init/update.
        Set skip_install to skip to installation process of the modules if not available.
        if skip_install is NOT set, the CancelOperation exception is raised if the installation failed."""
        if self.__git_resources is None:
            self.__git_resources = GitUtils(UserConfig().exegol_resources_path, "resources", "",
                                            skip_submodule_update=fast_load)
        if not self.__git_resources.isAvailable and not skip_install:
            self.__init_resources_repo()
        return self.__git_resources

    def __init_resources_repo(self):
        """Initialization procedure of exegol resources module.
        Raise CancelOperation if the initialization failed."""
        if ParametersManager().offline_mode:
            logger.error("It's not possible to install 'Exegol resources' in offline mode. Skipping the operation.")
            raise CancelOperation
        if Confirm("Do you want to download exegol resources? (~1G)", True):
            # If git wrapper is ready and exegol resources location is the corresponding submodule, running submodule update
            # if not, git clone resources
            if UserConfig().exegol_resources_path == ConstantConfig.src_root_path_obj / 'exegol-resources' and \
                    self.getWrapperGit().isAvailable:
                # When resources are load from git submodule, git objects are stored in the root .git directory
                self.__warningExcludeFolderAV(ConstantConfig.src_root_path_obj)
                if self.getWrapperGit().submoduleSourceUpdate("exegol-resources"):
                    self.__git_resources = None
                    self.getResourcesGit()
                else:
                    # Error during install, raise error to avoid update process
                    raise CancelOperation
            else:
                self.__warningExcludeFolderAV(UserConfig().exegol_resources_path)
                assert self.__git_resources is not None
                if not self.__git_resources.clone(ConstantConfig.EXEGOL_RESOURCES_REPO):
                    # Error during install, raise error to avoid update process
                    raise CancelOperation
        else:
            # User cancel installation, skip update update
            raise CancelOperation

    def isExegolResourcesReady(self) -> bool:
        """Update Exegol-resources from git (submodule)"""
        return self.getResourcesGit(fast_load=True).isAvailable

    @staticmethod
    def __warningExcludeFolderAV(directory: Union[str, Path]):
        """Generic procedure to warn the user that not antivirus compatible files will be downloaded and that
        the destination folder should be excluded from the scans to avoid any problems"""
        logger.warning(f"If you are using an [orange3][g]Anti-Virus[/g][/orange3] on your host, you should exclude the folder {directory} before starting the download.")
        while not Confirm(f"Are you ready to start the download?", True):
            pass



import os
import shutil
from typing import Optional, Dict, Sequence, Tuple

from docker.errors import NotFound, ImageNotFound
from docker.models.containers import Container

from exegol.console.ExegolPrompt import Confirm
from exegol.console.cli.ParametersManager import ParametersManager
from exegol.model.ContainerConfig import ContainerConfig
from exegol.model.ExegolContainerTemplate import ExegolContainerTemplate
from exegol.model.ExegolImage import ExegolImage
from exegol.model.SelectableInterface import SelectableInterface
from exegol.config.EnvInfo import EnvInfo
from exegol.utils.ExeLog import logger, console


class ExegolContainer(ExegolContainerTemplate, SelectableInterface):
    """Class of an exegol container already create in docker"""

    def __init__(self, docker_container: Container, model: Optional[ExegolContainerTemplate] = None):
        logger.debug(f"== Loading container : {docker_container.name}")
        self.__container: Container = docker_container
        self.__id: str = docker_container.id
        self.__xhost_applied = False
        if model is None:
            image_name = ""
            try:
                # Try to find the attached docker image
                docker_image = docker_container.image
            except ImageNotFound:
                # If it is not found, the user has probably forcibly deleted it manually
                logger.warning(f"Some images were forcibly removed by docker when they were used by existing containers!")
                logger.error(f"The '{docker_container.name}' containers might not work properly anymore and should also be deleted and recreated with a new image.")
                docker_image = None
                image_name = "[red bold]BROKEN[/red bold]"
            # Create Exegol container from an existing docker container
            super().__init__(docker_container.name,
                             config=ContainerConfig(docker_container),
                             image=ExegolImage(name=image_name, docker_image=docker_image),
                             hostname=docker_container.attrs.get('Config', {}).get('Hostname'))
            self.image.syncContainerData(docker_container)
            # At this stage, the container image object has an unknown status because no synchronization with a registry has been done.
            # This could be done afterwards (with container.image.autoLoad()) if necessary because it takes time.
            self.__new_container = False
        else:
            # Create Exegol container from a newly created docker container with its object template.
            super().__init__(docker_container.name,
                             config=ContainerConfig(docker_container),
                             # Rebuild config from docker object to update workspace path
                             image=model.image,
                             hostname=model.hostname)
            self.__new_container = True
        self.image.syncStatus()

    def __str__(self):
        """Default object text formatter, debug only"""
        return f"{self.getRawStatus()} - {super().__str__()}"

    def __getState(self) -> Dict:
        """Technical getter of the container status dict"""
        self.__container.reload()
        return self.__container.attrs.get("State", {})

    def getRawStatus(self) -> str:
        """Raw text getter of the container status"""
        return self.__getState().get("Status", "unknown")

    def getTextStatus(self) -> str:
        """Formatted text getter of the container status"""
        status = self.getRawStatus().lower()
        if status == "unknown":
            return "Unknown"
        elif status == "exited":
            return "[red]Stopped"
        elif status == "running":
            return "[green]Running"
        return status

    def isNew(self) -> bool:
        """Check if the container has just been created or not"""
        return self.__new_container

    def isRunning(self) -> bool:
        """Check is the container is running. Return bool."""
        return self.getRawStatus() == "running"

    def getFullId(self) -> str:
        """Container's id getter"""
        return self.__id

    def getId(self) -> str:
        """Container's short id getter"""
        return self.__container.short_id

    def getKey(self) -> str:
        """Universal unique key getter (from SelectableInterface)"""
        return self.name

    def start(self):
        """Start the docker container"""
        if not self.isRunning():
            logger.info(f"Starting container {self.name}")
            self.preStartSetup()
            with console.status(f"Waiting to start {self.name}", spinner_style="blue"):
                self.__container.start()

    def stop(self, timeout: int = 10):
        """Stop the docker container"""
        if self.isRunning():
            logger.info(f"Stopping container {self.name}")
            with console.status(f"Waiting to stop ({timeout}s timeout)", spinner_style="blue"):
                self.__container.stop(timeout=timeout)

    def spawnShell(self):
        """Spawn a shell on the docker container"""
        logger.info(f"Location of the exegol workspace on the host : {self.config.getHostWorkspacePath()}")
        for device in self.config.getDevices():
            logger.info(f"Shared host device: {device.split(':')[0]}")
        logger.success(f"Opening shell in Exegol '{self.name}'")
        # In case of multi-user environment, xhost must be set before opening each session to be sure
        self.__applyXhostACL()
        # Using system command to attach the shell to the user terminal (stdin / stdout / stderr)
        envs = self.config.getShellEnvs()
        options = ""
        if len(envs) > 0:
            options += f" -e {' -e '.join(envs)}"
        cmd = f"docker exec{options} -ti {self.getFullId()} {self.config.getShellCommand()}"
        logger.debug(f"Opening shell with: {cmd}")
        os.system(cmd)
        # Docker SDK doesn't support (yet) stdin properly
        # result = self.__container.exec_run(ParametersManager().shell, stdout=True, stderr=True, stdin=True, tty=True,
        #                                    environment=self.config.getShellEnvs())
        # logger.debug(result)

    def exec(self, command: Sequence[str], as_daemon: bool = True, quiet: bool = False, is_tmp: bool = False):
        """Execute a command / process on the docker container.
        Set as_daemon to not follow the command stream and detach the execution
        Set quiet to disable logs message
        Set is_tmp if the container will automatically be removed after execution"""
        if not self.isRunning():
            self.start()
        if not quiet:
            logger.info("Executing command on Exegol")
            if logger.getEffectiveLevel() > logger.VERBOSE and not ParametersManager().daemon:
                logger.info("Hint: use verbose mode to see command output (-v).")
        exec_payload, str_cmd = ExegolContainer.formatShellCommand(command, quiet)
        stream = self.__container.exec_run(exec_payload, environment={"CMD": str_cmd, "DISABLE_AUTO_UPDATE": "true"}, detach=as_daemon, stream=not as_daemon)
        if as_daemon and not quiet:
            logger.success("Command successfully executed in background")
        else:
            try:
                # stream[0] : exit code
                # stream[1] : text stream
                for log in stream[1]:
                    logger.raw(log.decode("utf-8"))
                if not quiet:
                    logger.success("End of the command")
            except KeyboardInterrupt:
                if not quiet and not is_tmp:
                    logger.info("Detaching process logging")
                    logger.warning("Exiting this command does [red]NOT[/red] stop the process in the container")

    @staticmethod
    def formatShellCommand(command: Sequence[str], quiet: bool = False, entrypoint_mode: bool = False) -> Tuple[str, str]:
        """Generic method to format a shell command and support zsh aliases.
        Set quiet to disable any logging here.
        Set entrypoint_mode to start the command with the entrypoint.sh config loader.
        - The first return argument is the payload to execute with every pre-routine for zsh.
        - The second return argument is the command itself in str format."""
        # Using base64 to escape special characters
        str_cmd = ' '.join(command)
        if not quiet:
            logger.success(f"Command received: {str_cmd}")
        # ZSH pre-routine: Load zsh aliases and call eval to force aliases interpretation
        cmd = f'autoload -Uz compinit; compinit; source ~/.zshrc; eval $CMD'
        if not entrypoint_mode:
            # For direct execution, the full command must be supplied not just the zsh argument
            cmd = f"zsh -c '{cmd}'"
        return cmd, str_cmd

    def remove(self):
        """Stop and remove the docker container"""
        self.__removeVolume()
        self.stop(timeout=2)
        logger.info(f"Removing container {self.name}")
        try:
            self.__container.remove()
            logger.success(f"Container {self.name} successfully removed.")
        except NotFound:
            logger.error(
                f"The container {self.name} has already been removed (probably created as a temporary container).")

    def __removeVolume(self):
        """Remove private workspace volume directory if exist"""
        volume_path = self.config.getPrivateVolumePath()
        # TODO add backup
        if volume_path != '':
            if volume_path.startswith('/wsl/') or volume_path.startswith('\\wsl\\'):
                # Docker volume defines from WSL don't return the real path, they cannot be automatically removed
                # TODO review WSL workspace volume
                logger.warning("Warning: WSL workspace directory cannot be removed automatically.")
                return
            logger.verbose("Removing workspace volume")
            logger.debug(f"Removing volume {volume_path}")
            try:
                list_files = os.listdir(volume_path)
            except PermissionError:
                if Confirm(f"Insufficient permission to view workspace files {volume_path}, "
                           f"do you still want to delete them?", default=False):
                    # Set list_files as empty to skip user prompt again
                    list_files = []
                else:
                    return
            except FileNotFoundError:
                logger.debug("This workspace has already been removed.")
                return
            try:
                if len(list_files) > 0:
                    # Directory is not empty
                    if not Confirm(f"Workspace [magenta]{volume_path}[/magenta] is not empty, do you want to delete it?",
                                   default=False):
                        # User can choose not to delete the workspace on the host
                        return
                # Try to remove files from the host with user permission (work only without sub-directory)
                shutil.rmtree(volume_path)
            except PermissionError:
                logger.info(f"Deleting the workspace files from the [green]{self.name}[/green] container as root")
                # If the host can't remove the container's file and folders, the rm command is exec from the container itself as root
                self.exec(["rm", "-rf", "/workspace"], as_daemon=False, quiet=True)
                try:
                    shutil.rmtree(volume_path)
                except PermissionError:
                    logger.warning(f"I don't have the rights to remove [magenta]{volume_path}[/magenta] (do it yourself)")
                    return
            except Exception as err:
                logger.error(err)
                return
            logger.success("Private workspace volume removed successfully")

    def preStartSetup(self):
        """
        Operation to be performed before starting a container
        :return:
        """
        self.__applyXhostACL()

    def postCreateSetup(self):
        """
        Operation to be performed after creating a container
        :return:
        """
        self.__applyXhostACL()

    def __applyXhostACL(self):
        """
        If GUI is enabled, allow X11 access on host ACL (if not already allowed) for linux and mac.
        On Windows host, WSLg X11 don't have xhost ACL.
        :return:
        """
        if self.config.isGUIEnable() and not self.__xhost_applied and not EnvInfo.isWindowsHost():
            self.__xhost_applied = True  # Can be applied only once per execution
            if shutil.which("xhost") is None:
                logger.error("The [green]xhost[/green] command is not available on your [bold]host[/bold]. "
                             "Exegol was unable to allow your container to access your graphical environment (or you don't have one).")
                return

            if EnvInfo.isMacHost():
                logger.debug(f"Adding xhost ACL to localhost")
                # add xquartz inet ACL
                with console.status(f"Starting XQuartz...", spinner_style="blue"):
                    os.system(f"xhost + localhost > /dev/null")
            else:
                logger.debug(f"Adding xhost ACL to local:{self.hostname}")
                # add linux local ACL
                os.system(f"xhost +local:{self.hostname} > /dev/null")



try:
    from exegol.manager.ExegolController import main
except ModuleNotFoundError as e:
    print("Mandatory dependencies are missing:", e)
    print("Please install them with pip3 install -r requirements.txt")
    exit(1)

if __name__ == "__main__":
    main()
